"""Handle and define plugins

Classes:

    - PluginWidget
    - PluginDialog
    - PluginSettingsWidget

Function to find and load plugins:
    - find_plugins

Plugin directory structure
--------------------------

Example:

    plugin_name
    ├── __init__.py
    ├── widgets.py
    └── dialogs.py

A plugin directory must be named in snake_case. It contains modules whose will
be described later and a __init__.py file which describes the plugin with the
following mandatory variables:

    .. code-block:: javascript

        __title__ = "PluginName"
        __description__ = "A plugin to cook variants"
        __author__ = "E. Micron"
        __version__ = "1.0.0"

Note:
    - `__title__` will be used on the GUI for the name of the plugin.
    - `__long_description__` can be used to provide high quality helper text via
        tooltip or What's This functionality (question mark button on top of all
        windows). It can be written in html.
        If it is not specified, the variable `__description__` will be used instead.

Plugin types
------------

Three kinds of plugins are supported, each one corresponds to a module in
a plugin directory.

    - Module named `widgets.py`:
        `widget` type: Parent class `PluginWidget`;
        Widget will be displayed on the main window according to the
        attribute `widget_location` of `PluginWidget`.
    - Module named `dialogs.py`:
        `dialog` type: Parent class `PluginDialog`;
        Dialog widget will be accessible from the tool menu of the UI.
    - Module named `settings.py`:
        `setting` type: Parent class `PluginSettingsWidget`;
        Widget will be accessible from the settings of the app.

That is, a plugin can support these 3 types simultaneously, but types must
be unique (i.e. no multiple widgets, or dialogs, or settings).


Module content
--------------

Instantiated classes of the modules will be those that inherit of the
expected parent class associated to their types (widget, dialog, setting).
In other words, each module **must** contains **one** class that inherits
from `PluginWidget` or `PluginDialog` or `PluginSettingsWidget`.

These classes **must** be named in accordance to the name of the plugin directory
followed by the corresponding type, in the CamelCase convention.

Example:

    A plugin `word_set` could contain one module named `widgets.py`.
    This module contains almost one class named `WordSetWidget` that inherits
    from `PluginWidget`.

"""
# Standard import
import os
import importlib
import pkgutil

# Qt imports
from PySide2.QtWidgets import QWidget, QDialog
from PySide2.QtCore import QSettings

# Cutevariant import
from cutevariant.gui import settings
import cutevariant.commons as cm

LOGGER = cm.logger()

DOCK_LOCATION = 1
CENTRAL_LOCATION = 2
FOOTER_LOCATION = 3


class PluginWidget(QWidget):
    """Model class for all widget plugins

    .. note:: Please override the functions of this class as much as possible.

    Class attributes:
        - LOCATION: Location of the plugin in the interface (QMainWindow)
            Can be: DOCK_LOCATION, CENTRAL_LOCATION, FOOTER_LOCATION
        - ENABLE: If False, the plugin is disabled and will not be loaded;
            (used for development purpose).
        - REFRESH_ONLY_VISIBLE: If this is True, the plugin will refresh only if plugin is visible

    Attributes:
        - mainwindow: Parent widget
        - widget_location: Instance variable, equivalent to class variable LOCATION
        - conn (sqlite3.connection): A connection to the sqlite project
        - dock (None, optional): Keep the attached dock to allow further clean
            deletion.
    """

    LOCATION = DOCK_LOCATION
    ENABLE = False
    REFRESH_ONLY_VISIBLE = True

    # This variable is set to False by the showEvent
    #  It is a hack to avoid calling load() 2 time by the showEvent in startup
    _STARTUP = True

    def __init__(self, parent=None):
        """Set parent window (mainwindow) to QWidget and to mainwindow attribute

        Keys:
            parent (QMainWindow): Mainwindow of Cutevariant, passed during
                plugin initialization.
        """
        super().__init__(parent)
        self.mainwindow = parent
        self.widget_location = DOCK_LOCATION
        self.conn = None
        self.dock = None

    def on_register(self, mainwindow):
        """Called when the mainwindow is build

        You should setup the mainwindow with your plugin from here.

        Don't forget to assign mainwindow attribute with the given argument
        if you didn't do it in the constructor. It's your last chance!

        Args:
            mainwindow (QMainWindow): cutevariant Mainwindow
        """
        pass

    def on_open_project(self, conn):
        """This method is called when a project is opened

        You should use the sql connector if your plugin uses the SQL database.

        Args:
            conn (sqlite3.connection): A connection to the sqlite project
        """
        pass

    def on_close(self):
        """Called when the mainwindow is closed **or** when the plugin is reset.

        In order to have a clean reset, please add routines to delete all
        widgets added to the mainwindow (actions in toolbars, etc.).

        Warnings:
            This routine **IS** important; if you override it, please don't forget
            to call `super().on_close()` at the end!

        .. seealso:: :meth:`cutevariant/gui/mainwindow.MainWindow.reset_ui`
        """
        self.close()
        self.deleteLater()
        LOGGER.debug("delete plugin... %s", self)
        if self.dock is not None:
            LOGGER.debug("delete its dock...")
            self.dock.close()
            self.dock.deleteLater()

    def on_refresh(self):
        """Called to refresh the GUI of the current plugin

        This is called by the mainwindow.controller::refresh methods
        """
        pass

    def showEvent(self, event):
        """Event called when a plugin is shown on the UI

        This is used to sync a plugin with the UI after it has been hidden.

        Note:
            Is also shown at the initialization => Test the SQL connection to
            avoid surprises due to an early call of `on_refresh`.

        Args:
            event(PySide2.QtGui.QShowEvent):
        """
        LOGGER.debug("Show event %s", self)

        if self.conn and not self._STARTUP:
            self.on_refresh()
        else:
            self._STARTUP = False

        # def create_settings(self):
        #     settings = QSettings()
        #     settings.beginGroup(self._prefix_settings)
        #     return settings

    @property
    def plugin_name(self):
        return cm.camel_to_snake(self.__class__.__name__.replace("Widget", ""))

    def create_settings(self):
        settings = QSettings()
        settings.beginGroup(self.plugin_name)
        return settings


class PluginDialog(QDialog):
    """Model class for all tool menu plugins

    These plugins are based on DialogBox; this means that they could be opened
    from the tools menu.
    """

    ENABLE = False

    def __init__(self, parent=None):
        """
        Keys:
            parent (QMainWindow): cutevariant Mainwindow
        """
        super().__init__(parent)
        self.conn = None

    @property
    def plugin_name(self):
        return cm.camel_to_snake(self.__class__.__name__.replace("Dialog", ""))

    def create_settings(self):
        settings = QSettings()
        settings.beginGroup(self.plugin_name)
        return settings


class PluginSettingsWidget(settings.SectionWidget):
    """Model class for settings plugins"""

    ENABLE = False

    def __init__(self, parent=None):
        """
        Keys:
            parent (QMainWindow): cutevariant window (mainly SettingsWidget)
        """
        super().__init__(parent)
        self.prefix_settings = self.plugin_name

    def on_refresh(self):
        """Called to refresh the GUI of the current plugin

        This is called by the mainwindow.controller::refresh methods
        """
        pass

    @property
    def plugin_name(self):
        return cm.camel_to_snake(self.__class__.__name__.replace("SettingsWidget", ""))


################################################################################


def find_plugins(path=None):
    """Find and return plugin descriptions from a directory

    See Also: Docstring of this module. For structure of a plugin directory.

    Example of yielded dict:


            {
                'name': 'word_set',
                'title': 'WordSet',
                'description': 'A plugin to manage word set',
                'long_description': 'Long description used with WhatsThis help',
                'version': '1.0.0',
                'widget': <class 'widgets.WordSetWidget'>,
                'dialog': <class 'widgets.PluginDialog'>
            }

    Keyword Arguments:
        path(str): Folder path where plugin are

    Returns:
        (generator[dict]): A dict with classes ready to be instantiated
    """
    # if path is None, return internal plugin path
    if path is None:
        plugin_path = os.path.join(os.path.dirname(__file__), "plugins")
    else:
        plugin_path = path

    # Loop over packages in plugins directory
    for package in pkgutil.iter_modules([plugin_path]):
        package_path = os.path.join(plugin_path, package.name)
        LOGGER.debug("Loading plugin at <%s>", package_path)

        spec = importlib.util.spec_from_file_location(
            package.name, os.path.join(package_path, "__init__.py")
        )

        # TODO: maybe could use __title__ to build class names...
        widget_class_name = cm.snake_to_camel(package.name) + "Widget"
        settings_class_name = cm.snake_to_camel(package.name) + "SettingsWidget"
        dialog_class_name = cm.snake_to_camel(package.name) + "Dialog"

        # Load __init__ file data of the module
        # We expect to load a plugin per module found in a plugin directory
        # This is the base dict of the item yielded from this function
        module = spec.loader.load_module()
        plugin_item = {
            "name": module.__name__,
            "title": module.__title__,
            "description": module.__description__,
            "long_description": module.__long_description__,
            "version": module.__version__,
        }

        authorized_module_classes = {
            "widgets": widget_class_name,
            "settings": settings_class_name,
            "dialogs": dialog_class_name,
        }

        authorized_base_clases = {
            "widgets": PluginWidget,
            "settings": PluginSettingsWidget,
            "dialogs": PluginDialog,
        }

        # Loop over modules in each plugin
        for sub_module_info in pkgutil.iter_modules([package_path]):
            LOGGER.debug("Loading module <%s>", sub_module_info)
            sub_module_type = sub_module_info.name

            # Filter module filenames
            if sub_module_type not in authorized_module_classes:
                continue

            # Dynamically load module
            sub_module_path = os.path.join(
                sub_module_info.module_finder.path, sub_module_type + ".py"
            )
            spec = importlib.util.spec_from_file_location(
                sub_module_type, sub_module_path
            )
            sub_module = spec.loader.load_module()

            # Filter not wanted classes (search main classes of the module)

            # Classes that don't have the module name in their name
            class_name = authorized_module_classes[sub_module_type]
            if class_name not in dir(sub_module):
                LOGGER.error(
                    "Plugin <%s.%s>, class <%s> not found!",
                    module.__name__,
                    sub_module_type,
                    class_name,
                )
                continue

            class_item = getattr(sub_module, class_name)
            # # Purge disabled plugins
            # if not class_item.ENABLE:
            #     LOGGER.debug(
            #         "Plugin <%s.%s> disabled",
            #         module.__name__,
            #         sub_module_type
            #     )
            #     continue

            # Classes that don't inherit of the expected Plugin class
            # See cutevariant/gui/plugin.py
            if authorized_base_clases[sub_module_type] not in class_item.__bases__:
                LOGGER.error(
                    "Plugin <%s.%s>, parent class <%s> not found!",
                    module.__name__,
                    sub_module_type,
                    authorized_base_clases[sub_module_type].__name__,
                )
                continue

            # Remove the "s" from module name...
            plugin_item[sub_module_type[:-1]] = class_item

            # Fix plugin status by user decision via app settings
            if not class_item.ENABLE:
                class_item.ENABLE = (
                    QSettings().value(f"plugins/{plugin_item['name']}/status") == "true"
                )

        yield plugin_item

    @property
    def plugin_name(self):
        return self.__class__.__name__.replace("SettingsWidget", "")


if __name__ == "__main__":

    print(cm.snake_to_camel("query_view"))



from PySide2.QtWidgets import QApplication

__title__ = "Columns selector"
__description__ = "A plugin to display fields "
__long_description__ = QApplication.instance().translate(
    "fields_editor",
    """
<p>This plugin is used to manage the fields visible from the main "variant view"
window.</p>

The fields are divided into 3 main categories:
<ul>
<li><b>variants:</b> fields concerning the variants in their globality
(chromosome, position, reference base, alternative base, etc.).</li>
<li><b>annotations:</b> fields added by annotation tools such as snpeff (gene,
consequence of the variant, impact rating, etc.).</li>
<li><b>samples:</b> fields concerning the individuals sequenced in the project.</li>
</ul>

<p>As the number of fields is large, it may be interesting to <em>use the keyword
search button</em> to quickly find those that are relevant in a study.</p>
""",
)
__author__ = "Sacha schutz"
__version__ = "1.0.0"



from PySide2.QtWidgets import QApplication

__title__ = "InfoVariant"
__description__ = "A plugin to display variant info"
__long_description__ = QApplication.instance().translate(
    "variant_info",
    """
<p>This plugin allows you to view all the data attached to the currently selected
variant in the main Variant view window.
</p>
<p>Where the main window and the fields editor selector apply to the variants as
a whole, this plugin targets only one variant.
</p>
<p>You will find the data of transcripts, samples, genotypes as well as manually
written comments added by right-clicking on the variant in the main window.
</p>
Data is read-only!
""",
)
__author__ = "Sacha schutz"
__version__ = "1.0.0"



"""Main window of Cutevariant"""
# Standard imports
import os
import sys
import sqlite3
from pkg_resources import parse_version
from functools import partial
from logging import DEBUG

# Qt imports
from PySide2.QtCore import Qt, QSettings, QByteArray, QDir, QUrl
from PySide2.QtWidgets import *
from PySide2.QtGui import QIcon, QKeySequence, QDesktopServices

# Custom imports
from cutevariant.core import get_sql_connection, get_metadatas, command
from cutevariant.core.writer import CsvWriter, PedWriter
from cutevariant.gui.ficon import FIcon
from cutevariant.gui.state import State
from cutevariant.gui.wizards import ProjectWizard
from cutevariant.gui.settings import SettingsDialog
from cutevariant.gui.widgets.aboutcutevariant import AboutCutevariant
from cutevariant import commons as cm
from cutevariant.commons import (
    MAX_RECENT_PROJECTS,
    DIR_ICONS,
    MIN_AUTHORIZED_DB_VERSION,
)

# Import plugins
from cutevariant.gui import plugin

LOGGER = cm.logger()


class MainWindow(QMainWindow):
    """Main window of Cutevariant"""

    def __init__(self, parent=None):

        super().__init__(parent)

        self.setWindowTitle("Cutevariant")
        self.toolbar = self.addToolBar("maintoolbar")
        self.toolbar.setObjectName("maintoolbar")  # For window saveState
        self.setWindowIcon(QIcon(DIR_ICONS + "app.png"))
        self.setWindowFlags(Qt.WindowContextHelpButtonHint | self.windowFlags())

        # Keep sqlite connection
        self.conn = None

        # App settings
        self.app_settings = QSettings()

        # State variable of application
        # store fields, source, filters, group_by, having data
        # Often changed by plugins
        self.state = State()

        # Central workspace
        self.central_tab = QTabWidget()
        self.footer_tab = QTabWidget()
        self.vsplit = QSplitter(Qt.Vertical)
        self.vsplit.addWidget(self.central_tab)
        self.vsplit.addWidget(self.footer_tab)
        self.setCentralWidget(self.vsplit)

        # Status Bar
        self.status_bar = QStatusBar()
        self.setStatusBar(self.status_bar)

        # Setup menubar
        self.setup_menubar()
        # Setup toobar under the menu bar
        self.setup_toolbar()

        # Register plugins
        self.plugins = {}  # dict of names (not titles) as keys and widgets as values
        self.dialog_plugins = {}  # dict of actions as keys and classes as values
        self.register_plugins()

        # Window geometry
        self.resize(600, 400)
        self.setGeometry(
            QApplication.instance().desktop().rect().adjusted(100, 100, -100, -100)
        )

        self.setTabPosition(Qt.AllDockWidgetAreas, QTabWidget.North)

        # If True, the GUI settings are deleted when the app is closed
        self.requested_reset_ui = False
        # Restores the state of this mainwindow's toolbars and dockwidgets
        self.read_settings()

        # Auto open recent projects
        recent = self.get_recent_projects()
        if recent and os.path.isfile(recent[0]):
            self.open(recent[0])

    def add_panel(self, widget, area=Qt.LeftDockWidgetArea):
        """Add given widget to a new QDockWidget and to view menu in menubar"""
        dock = QDockWidget()
        dock.setWindowTitle(widget.windowTitle())
        dock.setWidget(widget)
        dock.setStyleSheet("QDockWidget { font: bold }")
        # Keep the attached dock to allow further clean deletion
        widget.dock = dock
        # Set the objectName for a correct restoration after saveState
        dock.setObjectName(str(widget.__class__))

        self.addDockWidget(area, dock)
        self.view_menu.addAction(dock.toggleViewAction())

    def register_plugins(self):
        """Dynamically load plugins to the window

        Wrapper of :meth:`register_plugin`

        Two types of plugins can be registered:
        - widget: Added to the GUI,
        - dialog: DialogBox accessible from the tool menu.

        `setting` type is handled separately in :meth:`cutevariant.gui.settings.load_plugins`
        """
        LOGGER.info("MainWindow:: Registering plugins...")

        # Get classes of plugins
        # Don't forget to skip disabled plugins
        for extension in plugin.find_plugins():
            self.register_plugin(extension)

    def register_plugin(self, extension):
        """Dynamically load plugins to the window

        Two types of plugins can be registered:
        - widget: Added to the GUI,
        - dialog: DialogBox accessible from the tool menu.

        `setting` type is handled separately in :meth:`cutevariant.gui.settings.load_plugins`

        Args:
            extension (dict): Extension dict. See :meth:`cutevariant.gui.plugin.find_plugins`
        """
        LOGGER.debug("Extension: %s", extension)

        name = extension["name"]
        title = extension["title"]
        displayed_title = name if LOGGER.getEffectiveLevel() == DEBUG else title

        if "widget" in extension and extension["widget"].ENABLE:
            # New GUI widget
            plugin_widget_class = extension["widget"]

            # Setup new widget
            widget = plugin_widget_class(parent=self)
            if not widget.objectName():
                LOGGER.debug(
                    "widget '%s' has no objectName attribute; "
                    "=> fallback to extension name",
                    displayed_title,
                )
                widget.setObjectName(name)

            # Set title
            widget.setWindowTitle(displayed_title)

            # WhatsThis content
            long_description = extension.get("long_description")
            if not long_description:
                long_description = extension.get("description")
            widget.setWhatsThis(long_description)
            # Register (launch first init on some of them)
            widget.on_register(self)
            if self.conn:
                # If register occurs after a project is loaded we must set its
                # connection attribute
                widget.on_open_project(self.conn)

            # Init mainwindow via the constructor or on_register
            if widget.mainwindow != self:
                LOGGER.error(
                    "Bad plugin implementation, <mainwindow> plugin attribute is not set."
                )
                widget.on_close()
            else:
                # Add new plugin to plugins already registered
                self.plugins[name] = widget

                # Set position on the GUI
                if plugin_widget_class.LOCATION == plugin.DOCK_LOCATION:
                    self.add_panel(widget)

                if plugin_widget_class.LOCATION == plugin.CENTRAL_LOCATION:
                    self.central_tab.addTab(widget, widget.windowTitle())

                if plugin_widget_class.LOCATION == plugin.FOOTER_LOCATION:
                    self.footer_tab.addTab(widget, widget.windowTitle())

        if "dialog" in extension and extension["dialog"].ENABLE:
            # New tool menu entry
            plugin_dialog_class = extension["dialog"]

            # Add plugin to Tools menu
            dialog_action = self.tool_menu.addAction(displayed_title + "...")
            self.dialog_plugins[dialog_action] = plugin_dialog_class
            dialog_action.triggered.connect(self.show_dialog)

    def refresh_plugins(self, sender: plugin.PluginWidget = None):
        """Refresh all widget plugins

        It doesn't refresh the sender plugin, and not visible plugins.

        Args:
            sender (PluginWidget): from a plugin, you can pass "self" as argument
        """
        for plugin_obj in self.plugins.values():
            if plugin_obj is not sender and (
                plugin_obj.isVisible() or plugin_obj.REFRESH_ONLY_VISIBLE is False
            ):
                try:
                    plugin_obj.on_refresh()
                except Exception as e:
                    LOGGER.exception(e)

    def refresh_plugin(self, plugin_name: str):
        """Refresh a widget plugin identified by plugin_name

        It doesn't refresh the sender plugin

        Args:
            plugin_name (str): Name of the plugin to be refreshed
        """
        if plugin_name in self.plugins:
            plugin_obj = self.plugins[plugin_name]
            plugin_obj.on_refresh()

    def setup_menubar(self):
        """Menu bar setup: items and actions

        .. note:: Setup tools menu that could be dynamically augmented by plugins.
        """
        ## File Menu
        self.file_menu = self.menuBar().addMenu(self.tr("&File"))
        self.new_project_action = self.file_menu.addAction(
            FIcon(0xF01BA), self.tr("&New project"), self.new_project, QKeySequence.New
        )
        self.open_project_action = self.file_menu.addAction(
            FIcon(0xF095D),
            self.tr("&Open project..."),
            self.open_project,
            QKeySequence.Open,
        )
        ### Recent projects
        self.recent_files_menu = self.file_menu.addMenu(self.tr("Open recent"))
        self.setup_recent_menu()

        ### Export projects as
        self.export_csv_action = self.file_menu.addAction(
            self.tr("Export as csv"), self.export_as_csv
        )

        self.export_ped_action = self.file_menu.addAction(
            self.tr("Export pedigree PED/PLINK"), self.export_ped
        )

        self.file_menu.addSeparator()
        ### Misc
        self.file_menu.addAction(
            FIcon(0xF0493), self.tr("Settings..."), self.show_settings
        )
        self.file_menu.addSeparator()
        self.file_menu.addAction(self.tr("&Quit"), self.close, QKeySequence.Quit)

        ## Edit
        # TODO: if variant_view plugin is not loaded, disable this menu entries...
        self.edit_menu = self.menuBar().addMenu(self.tr("&Edit"))
        self.edit_menu.addAction(
            FIcon(0xF018F),
            self.tr("&Copy selected variants"),
            self.copy_variants_to_clipboard,
            "ctrl+shift+c",
        )
        self.edit_menu.addSeparator()
        self.edit_menu.addAction(
            FIcon(0xF0486),
            self.tr("&Select all variants"),
            self.select_all_variants,
            QKeySequence.SelectAll,
        )

        ## View
        self.view_menu = self.menuBar().addMenu(self.tr("&View"))
        self.view_menu.addAction(self.tr("Reset widgets positions"), self.reset_ui)
        # Set toggle footer visibility action
        show_action = self.view_menu.addAction(
            FIcon(0xF018D), self.tr("Show bottom toolbar")
        )
        show_action.setCheckable(True)
        show_action.setChecked(True)
        show_action.toggled.connect(self.toggle_footer_visibility)
        self.view_menu.addSeparator()

        ## Tools
        self.tool_menu = self.menuBar().addMenu(self.tr("&Tools"))

        ## Help
        self.help_menu = self.menuBar().addMenu(self.tr("Help"))

        self.help_menu.addAction(
            FIcon(0xF02D6),
            self.tr("What's this"),
            QWhatsThis.enterWhatsThisMode,
            QKeySequence.WhatsThis,
        )
        self.help_menu.addAction(
            FIcon(0xF059F),
            self.tr("Documentation..."),
            partial(QDesktopServices.openUrl, QUrl(cm.WIKI_URL, QUrl.TolerantMode)),
        )
        self.help_menu.addAction(
            FIcon(0xF0A30),
            self.tr("Report a bug..."),
            partial(
                QDesktopServices.openUrl, QUrl(cm.REPORT_BUG_URL, QUrl.TolerantMode)
            ),
        )

        self.help_menu.addSeparator()
        self.help_menu.addAction(
            self.tr("About Qt..."), QApplication.instance().aboutQt
        )
        self.help_menu.addAction(
            QIcon(DIR_ICONS + "app.png"),
            self.tr("About Cutevariant..."),
            self.about_cutevariant,
        )

    def setup_toolbar(self):
        """Top tool bar setup: items and actions

        .. note:: Require selection_widget and some actions of Menubar
        """
        # Tool bar
        self.toolbar.setToolButtonStyle(Qt.ToolButtonTextUnderIcon)
        self.toolbar.addAction(self.new_project_action)
        self.toolbar.addAction(self.open_project_action)
        self.toolbar.addAction(
            FIcon(0xF02D7), self.tr("Help"), QWhatsThis.enterWhatsThisMode
        )
        self.toolbar.addSeparator()

    def open(self, filepath):
        """Open the given db/project file

        .. note:: Called at the end of a project creation by the Wizard,
            and by Open/Open recent projects slots.

        :param filepath: Path of project file.
        :type filepath: <str>
        """
        if not os.path.isfile(filepath):
            return

        # Save directory
        self.app_settings.setValue("last_directory", os.path.dirname(filepath))

        # Create connection
        self.conn = get_sql_connection(filepath)

        try:
            # DB version filter
            db_version = get_metadatas(self.conn).get("cutevariant_version")
            if db_version and parse_version(db_version) < parse_version(
                MIN_AUTHORIZED_DB_VERSION
            ):
                # Refuse to open blacklisted DB versions
                # Unversioned files are still accepted
                QMessageBox.critical(
                    self,
                    self.tr("Error while opening project"),
                    self.tr("File: {} is too old; please create a new project.").format(
                        filepath
                    ),
                )
                return

            self.open_database(self.conn)
            self.save_recent_project(filepath)

        except sqlite3.OperationalError as e:
            LOGGER.exception(e)
            QMessageBox.critical(
                self,
                self.tr("Error while opening project"),
                self.tr("File: {}\nThe following exception occurred:\n{}").format(
                    filepath, e
                ),
            )
            return

        # Show the project name in title and in status bar
        self.setWindowTitle("Cutevariant - %s" % os.path.basename(filepath))
        self.status_bar.showMessage(self.tr("{} opened").format(filepath))

    def open_database(self, conn):
        """Open the project file and populate widgets

        Args:
            conn (sqlite3.Connection): Sqlite3 Connection
        """
        self.conn = conn

        # Clear memoization cache for count_cmd
        # Clear State variable of application
        # store fields, source, filters, group_by, having data
        self.state = State()

        for plugin_obj in self.plugins.values():
            plugin_obj.on_open_project(self.conn)

    def save_recent_project(self, path):
        """Save current project into QSettings

        Args:
            path (str): path of project
        """
        paths = self.get_recent_projects()
        if path in paths:
            paths.remove(path)
        paths.insert(0, path)

        self.app_settings.setValue("recent_projects", paths[:MAX_RECENT_PROJECTS])

    def get_recent_projects(self):
        """Return the list of recent projects stored in settings

        Returns:
            list: Recent project paths
        """
        # Reload last projects opened
        recent_projects = self.app_settings.value("recent_projects", list())

        # Check if recent_projects is a list() (as expected)
        if isinstance(recent_projects, str):
            recent_projects = [recent_projects]

        # Return only existing project files
        return [p for p in recent_projects if os.path.exists(p)]

    def clear_recent_projects(self):
        """Slot to clear the list of recent projects"""
        self.app_settings.remove("recent_projects")
        self.setup_recent_menu()

    def setup_recent_menu(self):
        """Setup recent submenu with previously opened projects"""
        self.recent_files_menu.clear()
        for path in self.get_recent_projects():
            self.recent_files_menu.addAction(path, self.on_recent_project_clicked)

        self.recent_files_menu.addSeparator()
        self.recent_files_menu.addAction(self.tr("Clear"), self.clear_recent_projects)

    def on_recent_project_clicked(self):
        """Slot to load a recent project"""
        action = self.sender()
        LOGGER.debug(action.text())
        self.open(action.text())

    def new_project(self):
        """Slot to allow creation of a project with the Wizard"""
        wizard = ProjectWizard()
        if not wizard.exec_():
            return

        db_filename = (
            wizard.field("project_path")
            + QDir.separator()
            + wizard.field("project_name")
            + ".db"
        )
        try:
            self.open(db_filename)
        except Exception as e:
            self.status_bar.showMessage(e.__class__.__name__ + ": " + str(e))
            raise

    def open_project(self):
        """Slot to open an already existing project from a QFileDialog"""
        # Reload last directory used
        last_directory = self.app_settings.value("last_directory", QDir.homePath())

        filepath, _ = QFileDialog.getOpenFileName(
            self,
            self.tr("Open project"),
            last_directory,
            self.tr("Cutevariant project (*.db)"),
        )
        if filepath:
            try:
                self.open(filepath)
            except Exception as e:
                self.status_bar.showMessage(e.__class__.__name__ + ": " + str(e))
                raise

    def export_as_csv(self):
        """Export variants into CSV file"""
        # Reload last directory used
        last_directory = self.app_settings.value("last_directory", QDir.homePath())

        filepath, _ = QFileDialog.getSaveFileName(
            self, self.tr("Save project"), last_directory, self.tr("(*.csv)")
        )

        if filepath:
            with open(filepath, "w") as file:
                writer = CsvWriter(file)
                writer.save(self.conn)

    def export_ped(self):
        """Export samples into PED/PLINK file"""
        # Reload last directory used
        last_directory = self.app_settings.value("last_directory", QDir.homePath())

        # noinspection PyCallByClass
        filepath, _ = QFileDialog.getSaveFileName(
            self, self.tr("Save project"), last_directory, "(*.tfam)"
        )

        if filepath:
            filepath = filepath if filepath.endswith(".tfam") else filepath + ".tfam"

            with open(filepath, "w") as file:
                writer = PedWriter(file)
                writer.save(self.conn)

    def show_settings(self):
        """Slot to show settings window"""
        widget = SettingsDialog(self)
        widget.uiSettingsChanged.connect(self.reload_ui)
        widget.exec_()

    def show_dialog(self):
        """Show Plugin dialog box after a click on the tool menu"""
        action = self.sender()
        if action in self.dialog_plugins:
            # Get class object and instantiate it
            dialog_class = self.dialog_plugins[action]
            # Send SQL connection
            dialog = dialog_class(conn=self.conn)
            dialog.mainwindow = self
            dialog.exec_()

    def about_cutevariant(self):
        """Slot to show about window"""
        dialog_window = AboutCutevariant(self)
        dialog_window.exec_()

    def reload_ui(self):
        """Reload *without* reset the positions of the widgets"""
        geometry = self.saveGeometry()
        ui_state = self.saveState()
        self.reset_ui()
        # Reload positions
        self.restoreGeometry(geometry)
        self.restoreState(ui_state)

    def reset_ui(self):
        """Reset the positions of docks (and their widgets) to the default state

        All the widgets are deleted and reinstantiated on the GUI.
        GUI settings via QSettings are also reset.
        """
        # Set reset ui boolean (used by closeEvent)
        self.requested_reset_ui = True
        # Reset settings
        self.write_settings()

        # Remove widgets in QTabWidgets
        for index in range(self.central_tab.count()):
            self.central_tab.removeTab(index)

        for index in range(self.footer_tab.count()):
            self.footer_tab.removeTab(index)

        # Remove tool menu actions linked to the dialog plugins
        for action in self.dialog_plugins:
            # LOGGER.debug("Remove action <%s>", action.text())
            self.tool_menu.removeAction(action)
            action.deleteLater()

        # Purge widgets (central/footer and others) and related docks
        for plugin_obj in self.plugins.values():
            # LOGGER.debug("Remove plugin <%s>", plugin_obj)
            self.removeDockWidget(plugin_obj.dock)
            plugin_obj.on_close()

        # Clean registered plugins
        self.plugins = {}
        self.dialog_plugins = {}
        # Register new plugins
        self.register_plugins()
        self.open_database(self.conn)
        # Allow a user to save further modifications
        self.requested_reset_ui = False

    def deregister_plugin(self, extension):
        """Delete plugin from the UI; Called from app settings when a plugin is disabled.

        - dialogs plugins: Remove action from tool menu
        - widgets plugins: Delete widget and its dock if available via its
            on_close() method.

        Args:
            extension (dict): Extension dict. See :meth:`cutevariant.gui.plugin.find_plugins`
        """
        name = extension["name"]
        title = extension["title"]
        displayed_title = name if LOGGER.getEffectiveLevel() == DEBUG else title

        # Remove tool menu actions linked to the dialog plugins
        for action in self.dialog_plugins:
            if action.text() == displayed_title:
                # LOGGER.debug("Remove action <%s>", action.text())
                self.tool_menu.removeAction(action)
                action.deleteLater()
                del self.dialog_plugins[action]
                break

        # Purge widgets and related docks
        if name in self.plugins:
            plugin_obj = self.plugins[name]
            # LOGGER.debug("Remove plugin <%s>", plugin_obj)
            self.removeDockWidget(plugin_obj.dock)
            plugin_obj.on_close()
            del self.plugins[name]

    def copy_variants_to_clipboard(self):
        if "variant_view" in self.plugins:
            self.plugins["variant_view"].copy()

    def select_all_variants(self):
        """Select all elements in the current tab's view"""
        if "variant_view" in self.plugins:
            self.plugins["variant_view"].select_all()

    def closeEvent(self, event):
        """Save the current state of this mainwindow's toolbars and dockwidgets

        .. warning:: Make sure that the property objectName is unique for each
            QToolBar and QDockWidget added to the QMainWindow.

        .. note:: Reset windowState if asked.
        """
        self.write_settings()
        super().closeEvent(event)

    def write_settings(self):
        """Store the state of this mainwindow.

        .. note:: This methods is called by closeEvent
        """
        if self.requested_reset_ui:
            # Delete window state setting
            self.app_settings.remove("windowState")
        else:
            self.app_settings.setValue("geometry", self.saveGeometry())
            #  TODO: handle UI changes by passing UI_VERSION to saveState()
            self.app_settings.setValue("windowState", self.saveState())

    def read_settings(self):
        """Restore the state of this mainwindow's toolbars and dockwidgets

        .. note:: If windowState is not stored, current state is written.
        """
        # Init reset ui boolean
        self.requested_reset_ui = False

        self.restoreGeometry(QByteArray(self.app_settings.value("geometry")))
        #  TODO: handle UI changes by passing UI_VERSION to saveState()
        window_state = self.app_settings.value("windowState")
        if window_state:
            self.restoreState(QByteArray(window_state))
        else:
            # Setting has been deleted: set the current default state
            #  TODO: handle UI changes by passing UI_VERSION to saveState()
            self.app_settings.setValue("windowState", self.saveState())

    def toggle_footer_visibility(self, visibility):
        """Toggle visibility of the bottom toolbar and all its plugins"""
        self.footer_tab.setVisible(visibility)

    # @Slot()
    # def on_query_model_changed(self):
    #     for name, _plugin in self.plugins.items():
    #         if _plugin.isVisible():
    #             _plugin.on_query_model_changed(self.query_model)

    # @Slot()
    # def on_variant_changed(self, variant):
    #     for name, _plugin in self.plugins.items():
    #         if _plugin.isVisible():
    #             _plugin.on_variant_changed(variant)


if __name__ == "__main__":
    app = QApplication(sys.argv)

    w = MainWindow()

    w.show()

    app.exec_()



# Standard imports
import sqlite3

# Qt imports
from PySide2.QtWidgets import (
    QVBoxLayout,
    QFormLayout,
    QTableView,
    QComboBox,
    QGroupBox,
    QLabel,
    QSizePolicy,
    QSpacerItem,
    QDialogButtonBox,
)
from PySide2.QtCore import Qt, QAbstractTableModel, QModelIndex, QThreadPool

# Custom imports
from cutevariant.gui.ficon import FIcon
from cutevariant.gui.plugin import PluginDialog
from cutevariant.gui.sql_thread import SqlThread
from cutevariant.core import sql


class TrioAnalysisDialog(PluginDialog):

    ENABLE = True

    DE_NOVO = 0
    AUTOSOMAL_RECESSIVE = 1
    AUTOSOMAL_DOMINANT = 2
    X_LINKED_RECESSIVE = 3

    def __init__(self, conn=None, parent=None):
        super().__init__(parent)
        self.conn = conn

        self.type_combo = QComboBox()

        self.type_combo.addItem(self.tr("De novo mutation"), self.DE_NOVO)
        self.type_combo.addItem(
            self.tr("Autosomal Recessive"), self.AUTOSOMAL_RECESSIVE
        )
        self.type_combo.addItem(self.tr("Autosomal Dominant"), self.AUTOSOMAL_DOMINANT)
        self.type_combo.addItem(self.tr("X-linked Recessive"), self.X_LINKED_RECESSIVE)

        self.form_box = QGroupBox()
        self.father_combo = QComboBox()
        self.mother_combo = QComboBox()
        self.child_combo = QComboBox()

        # Connect to check form
        self.type_combo.currentTextChanged.connect(self.check_form)
        self.father_combo.currentTextChanged.connect(self.check_form)
        self.mother_combo.currentTextChanged.connect(self.check_form)
        self.child_combo.currentTextChanged.connect(self.check_form)

        flayout = QFormLayout()
        flayout.addRow(self.tr("Type"), self.type_combo)

        # add spacer

        # flayout.addItem(QSpacerItem(0, 10))
        flayout.addRow(self.tr("Father"), self.father_combo)
        flayout.addRow(self.tr("Mother"), self.mother_combo)
        flayout.addRow(self.tr("Child"), self.child_combo)

        self.form_box.setLayout(flayout)

        self.button_box = QDialogButtonBox(
            QDialogButtonBox.Apply | QDialogButtonBox.Cancel
        )
        self.button_box.button(QDialogButtonBox.Apply).setDisabled(True)
        self.button_box.button(QDialogButtonBox.Apply).clicked.connect(
            self.create_filter
        )
        self.button_box.rejected.connect(self.reject)

        vlayout = QVBoxLayout()

        self.title = QLabel()
        self.title.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed)
        self.title.setText(
            """<b>Create Trio analysis filters</b> <br/>
            Set samples to the corresponding parents
            """
        )
        vlayout.addWidget(self.title)
        vlayout.addWidget(self.form_box)
        vlayout.addWidget(self.button_box)
        self.setLayout(vlayout)
        self.resize(600, 300)
        self.populate()

    def populate(self):
        """Fill combobox with samples from databases 
        """

        self.mother_combo.clear()
        self.father_combo.clear()
        self.child_combo.clear()

        samples = [i["name"] for i in sql.get_samples(self.conn)]

        for sample in samples:
            self.mother_combo.addItem(FIcon(0xF1077), sample)
            self.father_combo.addItem(FIcon(0xF0643), sample)
            self.child_combo.addItem(FIcon(0xF0E7D), sample)

    def check_form(self):
        """ Check if formular is okay and enable apply button
        This methods is triggered by all formular input 
        """

        #  Check if all samples are unique
        valid_form = False
        samples = {
            self.mother_combo.currentText(),
            self.father_combo.currentText(),
            self.child_combo.currentText(),
        }

        if len(samples) == 3:
            valid_form = True

        self.button_box.button(QDialogButtonBox.Apply).setEnabled(valid_form)

    def create_filter(self):
        """ build filter and send to the mainwindow.state """

        filter_type = self.type_combo.currentData()

        father = self.father_combo.currentText()
        mother = self.mother_combo.currentText()
        child = self.child_combo.currentText()

        if filter_type == self.DE_NOVO:
            filters = {
                "AND": [
                    {"field": ("sample", father, "gt"), "operator": "=", "value": 0},
                    {"field": ("sample", mother, "gt"), "operator": "=", "value": 0},
                    {"field": ("sample", child, "gt"), "operator": "=", "value": 1},
                ]
            }

        elif filter_type == self.AUTOSOMAL_RECESSIVE:
            filters = {
                "AND": [
                    {"field": ("sample", father, "gt"), "operator": "=", "value": 1},
                    {"field": ("sample", mother, "gt"), "operator": "=", "value": 1},
                    {"field": ("sample", child, "gt"), "operator": "=", "value": 2},
                ]
            }

        elif filter_type == self.AUTOSOMAL_DOMINANT:
            filters = {
                "AND": [
                    {"field": ("sample", child, "gt"), "operator": "=", "value": 1.0},
                    {
                        "OR": [
                            {
                                "AND": [
                                    {
                                        "field": ("sample", father, "gt"),
                                        "operator": "=",
                                        "value": 1.0,
                                    },
                                    {
                                        "field": ("sample", mother, "gt"),
                                        "operator": "=",
                                        "value": 0.0,
                                    },
                                ]
                            },
                            {
                                "AND": [
                                    {
                                        "field": ("sample", father, "gt"),
                                        "operator": "=",
                                        "value": 0.0,
                                    },
                                    {
                                        "field": ("sample", mother, "gt"),
                                        "operator": "=",
                                        "value": 1.0,
                                    },
                                ]
                            },
                        ]
                    },
                ]
            }

        elif filter_type == self.X_LINKED_RECESSIVE:
            filters = {
                "AND": [
                    {"field": "chr", "operator": "=", "value": "X"},
                    {"field": ("sample", father, "gt"), "operator": "=", "value": 0},
                    {"field": ("sample", mother, "gt"), "operator": "=", "value": 1},
                    {"field": ("sample", child, "gt"), "operator": ">", "value": 0},
                ]
            }

        self.mainwindow.state.filters = filters
        self.mainwindow.refresh_plugins()
        self.close()


if __name__ == "__main__":
    from PySide2.QtWidgets import QApplication
    import sys

    app = QApplication(sys.argv)
    conn = sql.get_sql_connection("/home/sacha/Dev/corpasome.db")
    dialog = TrioAnalysisDialog(conn=conn)
    dialog.show()
    app.exec_()



# Standard imports
import sqlite3
from typing import Callable

# Qt imports
from PySide2.QtCore import QThread, QObject, Signal

# Custom imports
from cutevariant.core.sql import get_sql_connection
from cutevariant.commons import logger

LOGGER = logger()


class SqlThread(QThread):
    """Used to execute SQL/VQL queries in a separated thread

    Attributes:
        - db_file (str): File path of the database.
        - async_conn (sqlite3.Connection): sqlite3 Connection
        - function (Callable): Function to be executed
        - caching_hash: A user defined hash of the request
        - results: Contain the result of the threaded function.
            `None`, as long as the function has not finished its execution done.



    Signals:
        - result_ready(): Emitted when results is available. If no results or error ,
            This signal is not emitted
        - error(str): Emitted when the function has encountered an error during
            its execution. The message is formatted with the type and the
            message of the exception.
        - finished() : inherits from QThread
        - starte() : inherits from QThread


    Notes:
        AutoDelete flag of such objects is set to False.
        This means that QThreadPool **will not** delete these objects after
        calling run() method.
    """

    error = Signal(str)
    result_ready = Signal()

    def __init__(self, conn: sqlite3.Connection = None, function: Callable = None):
        """Init a Thread with sqlite connection and callable

        Notes:
            Since sqlite3 Connection objects are not thread safe, we create a new
            connection based on it. 



        Examples:

            >>> thread = SqlThread(conn)
            >>> thread.start_function(lambda conn: conn.execute("SELECT COUNT(*) FROM variants"))
            >>> thread.result_ready.connect(self.on_received)



        Args:
            conn (sqlite3.Connection): sqlite3 Connexion

        """

        super().__init__()

        self.conn = conn
        self._async_conn = None
        self.results = None
        self.function = function
        self.last_error = None

    @property
    def conn(self) -> sqlite3.Connection:
        """Return the Application thread connection
        
        Returns:
            sqlite3.Connection
        """
        return self._conn

    @property
    def async_conn(self) -> sqlite3.Connection:
        """Return the thread connection 
        It is a clone of application connection for thread safe
        
        Returns:
            TYPE: Description
        """
        return self._async_conn

    @conn.setter
    def conn(self, conn):
        if conn:
            self._conn = conn
            self._async_conn = None
            self.db_file = conn.execute("PRAGMA database_list").fetchone()["file"]

    def run(self):
        """Execute the function in a new thread

        Note:
            All the context of this function is executed into a separated thread.

        """
        self.last_error = None

        if self.function is None:
            LOGGER.exception("no function defined")
            return

        self._async_conn = get_sql_connection(self.db_file)
        assert self.async_conn

        try:
            LOGGER.debug("thread start ")
            self.results = self.function(self.async_conn)
            LOGGER.debug("Thread finished")
        except Exception as e:
            # LOGGER.exception(e)
            self.last_error = "%s: %s" % (e.__class__.__name__, str(e))
            self.async_conn.close()
            self.error.emit(self.last_error)
        else:
            self.async_conn.close()
            self.result_ready.emit()

        return

    def start_function(self, function: Callable, caching_hash=None):
        """Execute a function in the thread
        
        Args:
            function (Callable): Function must take con as arguments 

        Examples:
            >>> thread.exec_function(lambda conn: conn.execute("SELECT ..."))

        """
        assert isinstance(function, Callable)
        self.function = function
        self.start()

    def interrupt(self):
        """Interrupt the thread connection

        Use this method instead of terminate.
        """
        if self.async_conn:
            self.async_conn.interrupt()

    @property
    def function(self):
        return self._function

    @function.setter
    def function(self, value: Callable):
        # A valid function must be set
        self._function = value


if __name__ == "__main__":
    from PySide2.QtWidgets import QApplication
    import sys

    app = QApplication(sys.argv)

    app.exec_()



"""Module to handle BED files"""
# Standard imports
import gzip
import csv
import os
import io
import re

# Custom imports
import cutevariant.commons as cm

LOGGER = cm.logger()


class BedReader:
    r"""BED file parser

    It is a substitution to pybedtools that is a (too) big black box to make lots
    of things, while we only need the file parser.
    Thus, pybedtools is not multiplatform compatible (thanks to Winshit OS).
    Thus (again), pybedtools doesn't support files with spaces as separators
    (while the specification is implicit about it)...

    We support:
        - Simple text files
        - BED data as string
        - Compressed gzip files
        - Tabs and spaces files
        - Bedgraph files
        - BED files with headers

    .. seealso:: BED specs: https://www.ensembl.org/info/website/upload/bed.html

    How to use::

        intervals = BedReader("myfile.bed.gz)
        generator = iter(intervals)
        first_interval = next(generator)

        intervals = BedReader("myfile.bed)

        large_string = \"""
            chr1 1    10   feature1  0 +
            chr1 50   60   feature2  0 -
            chr1 51 59 another_feature 0 +
        \"""
        intervals = BedReader(large_string)

        for interval in intervals:
            print(interval)
    """

    def __init__(self, filepath, *args, **kwargs):

        self.count = 0
        self.filepath = filepath

        # Autodetection of string and filepaths
        if not os.path.exists(self.filepath):
            self.is_from_string = True
            self.is_gz_file = False
        else:
            self.is_from_string = False
            self.is_gz_file = cm.is_gz_file(filepath)

    def __iter__(self):
        """Yield Interval objects in the given BED file

        Each Interval object is an OrderedDict with the following keys:
            - chrom
            - start
            - end
            - name
            - score
            - strand
            - thickStart
            - thickEnd
            - itemRgb
            - blockCount
            - blockSizes
            - blockStarts

        Excedent data is put in and additional column 'misc'.
        Empty columns contain None values.

        :return: Generator of Intervals
        :rtype: <generator <OrderectDict>>
        """
        if self.is_from_string:
            # Clean the string:
            # Remove spaces at the start and end
            # Remove duplicated spaces in the string

            self.filepath = re.sub(" +", " ", self.filepath.strip())
            self.filepath = re.sub("\n ", "\n", self.filepath.strip())

            # Load in memory file
            yield from self.get_intervals(io.StringIO(self.filepath))
        else:
            if self.is_gz_file:
                # Handle gzip file
                with gzip.open(self.filepath, "rt") as stream:
                    yield from self.get_intervals(stream)
            else:
                # Handle text file
                with open(self.filepath, "r") as stream:
                    yield from self.get_intervals(stream)

    def get_intervals(self, stream):
        """Yield Interval objects in the given stream

        .. seealso:: :meth:`__iter__`
        """
        # Throws line with headers
        skipped_header_line = 0  # Will be used to rewind the stream
        for line in stream:
            if line.startswith(("@", "#", "track", "browser")) or not line.strip():
                # Header detected
                LOGGER.debug("comment: %s", line)
                skipped_header_line += 1
                continue
            break

        # Quick tests on the first line of data...
        # Delimiters can only be '\t' or ' ' since
        # 'itemRgb' column is comma separated.

        # Rewind the stream
        stream.seek(0)
        [next(stream) for _ in range(skipped_header_line)]
        try:
            # If there is no data at all or after the header
            data_line = next(stream)
        except StopIteration:
            LOGGER.debug("No interval detected in the given BED file!")
            return

        csv_dialect = csv.Sniffer().sniff(data_line, delimiters="\t ")

        # Rewind the stream
        stream.seek(0)
        [next(stream) for _ in range(skipped_header_line)]
        # Build a csv reader
        bed_fieldnames = (
            "chrom",
            "start",
            "end",
            "name",
            "score",
            "strand",
            "thickStart",
            "thickEnd",
            "itemRgb",
            "blockCount",
            "blockSizes",
            "blockStarts",
        )
        csv_reader = csv.DictReader(
            stream, fieldnames=bed_fieldnames, restkey="misc", dialect=csv_dialect
        )

        line_number = 0
        for line_number, interval in enumerate(csv_reader, 1):
            # print(interval)
            yield interval

        self.count = line_number


def parse_bed_file(filepath):
    """Parse the given BED file with 'pybedtools' package, yield features in the file

    .. note:: chromosom field is processed to remove 'chr' prefixes.

    .. note:: Intervals yielded have the following attributes:
        interval.chrom, interval.start, interval.stop

        Other attributes are described here:
        https://daler.github.io/pybedtools/intervals.html

    We assume that the given BED file is composed of lines with the following fields:
        - chromosom
        - start position
        - end position

    .. seealso:: https://www.ensembl.org/info/website/upload/bed.html
    """
    bedtool = BedReader(filepath)

    for interval in bedtool:
        # Remove 'chr' prefix from the chromosome name
        # interval["chrom"] = interval["chrom"].replace("chr", "")

        yield interval



# Standard imports
import argparse
import os
import sys
from functools import partial

# Custom imports
import progressbar
from columnar import columnar
from cutevariant.core.importer import async_import_file
from cutevariant.core import sql, vql, command
from cutevariant.core.querybuilder import *
from cutevariant.commons import log_level


def display_sql_results(data, headers, *args, **kwargs):
    """Display SQL results in the console in tabulated format"""
    print(columnar(data, headers=headers, no_borders=True, **kwargs,))


def display_query_status(query_result):
    """Display error status of a valid and executed VQL query

    Supported results of VQL commands: drop/import; create/set/bed

    Args:
        query_result(dict)
    """
    # Success for drop/import; id present for create/set/bed
    if query_result.get("success") or ("id" in query_result):
        print("Done")
    else:
        print("An error occured, there is no result, or there is nothing to do.")
    exit(1)


def main():
    # noinspection PyTypeChecker
    parser = argparse.ArgumentParser(
        formatter_class=lambda prog: argparse.RawDescriptionHelpFormatter(prog),
        description="""
Cutevariant cli mode helps to run actions directly from command-line.\n
The env variable $CUTEVARIANT_DB can be used to define a database instead of
the arguments.""",
        epilog="""Examples:

    $ cutevariant-cli show --db my_database.db samples
    or
    $ export CUTEVARIANT_DB=my_database.db
    $ cutevariant-cli show samples""",
    )
    # Default log level: critical
    parser.add_argument(
        "-vv",
        "--verbose",
        nargs="?",
        default="error",
        choices=["debug", "info", "critical", "error", "warning"],
    )

    sub_parser = parser.add_subparsers(dest="subparser")

    # Common parser: Database file requirement #################################
    parent_parser = argparse.ArgumentParser(add_help=False)
    parent_parser.add_argument(
        "--db", help="SQLite database. By default, $CUTEVARIANT_DB is used."
    )

    # Create DB parser #########################################################
    createdb_parser = sub_parser.add_parser(
        "createdb",
        help="Build a SQLite database from a vcf file",
        parents=[parent_parser],
        epilog="""Examples:

        $ cutevariant-cli createdb -i "examples/test.snpeff.vcf"
        """,
    )
    createdb_parser.add_argument("-i", "--input", help="VCF file path", required=True)

    # Show parser ##############################################################
    show_parser = sub_parser.add_parser(
        "show", help="Display table content", parents=[parent_parser]
    )
    show_parser.add_argument(
        "table",
        choices=["fields", "selections", "samples", "wordsets"],
        help="Possible names of tables.",
    )

    # Remove parser ############################################################
    remove_parser = sub_parser.add_parser(
        "remove", help="remove selection", parents=[parent_parser]
    )
    remove_parser.add_argument("names", nargs="+", help="Name(s) of selection(s).")

    # VQL parser ###############################################################
    select_parser = sub_parser.add_parser(
        "exec",
        help="Execute a VQL statement.",
        parents=[parent_parser],
        formatter_class=argparse.RawDescriptionHelpFormatter,
        epilog="""Examples:

    $ cutevariant-cli exec "SELECT favorite,chr,pos,ref,alt FROM variants"
    or
    $ cutevariant-cli exec "SELECT chr,ref,alt FROM variants" -s myselection
    or
    $ cutevariant-cli exec "IMPORT WORDSETs 'examples/gene.txt' AS mygenes"
    or
    $ cutevariant-cli exec "DROP WORDSETS mygenes"
    or
    $ cutevariant-cli exec "CREATE myselection1 FROM variants WHERE gene = 'CHID1'"
    $ cutevariant-cli exec "CREATE myselection2 FROM variants WHERE gene = 'CICP23'"
    $ cutevariant-cli exec "CREATE myselection3 = myselection2 | myselection2"
    or
    $ cutevariant-cli exec "CREATE boby FROM variants INTERSECT 'examples/test.bed'"
    """,
    )
    select_parser.add_argument("vql", help="A VQL statement.")
    select_parser.add_argument(
        "-l",
        "--limit",
        help="Limit the number of lines in output.",
        type=int,
        default=100,
    )
    # select_parser.add_argument(
    #     "-g",
    #     "--group",
    #     action="store_true",
    #     help="Group SELECT query by...(chr,pos,ref,alt).",
    # )
    select_parser.add_argument(
        "-s", "--to-selection", help="Save SELECT query into a selection name."
    )

    # Set parser ###############################################################
    # set_parser = sub_parser.add_parser("set", help="Set variable", parents=[parent_parser])

    # Workaround for sphinx-argparse module that require the object parser
    # before the call of parse_args()
    if "html" in sys.argv:
        return parser

    args = parser.parse_args()

    if len(sys.argv) == 1:
        parser.print_help(sys.stderr)
        sys.exit(1)

    # Set log level
    # _logger.setLevel(logging.DEBUG)
    log_level(args.verbose)

    # Create DB parser #########################################################
    if args.subparser == "createdb":
        if not args.db:
            # Database file is not set:
            # The output file will be based on the name of the VCF one
            args.db = args.input + ".db"

        if os.path.exists(args.db):
            # Remove existing file
            os.remove(args.db)

        conn = sql.get_sql_connection(args.db)
        if conn:
            # TODO: bug ... max is not 100...
            for i, message in progressbar.progressbar(
                async_import_file(conn, args.input), redirect_stdout=True
            ):
                print(message)

        print("The database is successfully created!")
        exit()

    # Prepare SQL connection on DB file
    if "CUTEVARIANT_DB" in os.environ and args.subparser != "createdb":
        args.db = os.environ["CUTEVARIANT_DB"]
    elif not args.db:
        print("You must specify a database file via $CUTEVARIANT_DB or --db argument")
        exit()

    # Init SQL connection
    conn = sql.get_sql_connection(args.db)

    # Show parser ##############################################################
    if args.subparser == "show":
        if args.table == "fields":
            display_sql_results(
                (i.values() for i in sql.get_fields(conn)),
                ["id", "name", "table", "type", "description"],
            )

        if args.table == "samples":
            display_sql_results(
                (i.values() for i in sql.get_samples(conn)), ["id", "name"]
            )

        if args.table == "selections":
            display_sql_results(
                (i.values() for i in sql.get_selections(conn)),
                ["id", "name", "variant_count"],
            )

        if args.table == "wordsets":
            display_sql_results(
                (i.values() for i in sql.get_wordsets(conn)), ["id", "word_count"]
            )

    # Remove parser ############################################################
    if args.subparser == "remove":
        for name in args.names:
            sql.delete_selection_by_name(conn, name)

    # VQL parser ###############################################################
    if args.subparser == "exec":
        query = "".join(args.vql)
        vql_command = None

        # Test the VQL query
        try:
            cmd = vql.parse_one_vql(query)
        except (vql.textx.TextXSyntaxError, vql.VQLSyntaxError) as e:
            # Available attributes: e.message, e.line, e.col
            print("%s: %s, col: %d" % (e.__class__.__name__, e.message, e.col))
            print("For query:", query)
            exit(0)

        # Select command with redirection to selection
        if cmd["cmd"] == "select_cmd" and args.to_selection:
            vql_command = partial(
                command.create_cmd,
                conn,
                args.to_selection,
                source=cmd["source"],
                filters=cmd["filters"],
            )

        try:
            # Is it redundant with check_vql ?
            # No because we also execute SQL statement here
            if vql_command:
                ret = vql_command()
            else:
                ret = command.create_command_from_obj(conn, cmd)()
            if not isinstance(ret, dict):
                # For drop_cmd, import_cmd,
                ret = list(ret)
        except (sqlite3.DatabaseError, vql.VQLSyntaxError) as e:
            LOGGER.exception(e)
            exit(0)

        LOGGER.debug("SQL result: %s", ret)
        LOGGER.debug("VQL command: %s", cmd["cmd"])
        # Note: show_cmd is supported in a separated command option

        # Select command
        if cmd["cmd"] in ("select_cmd",) and not args.to_selection:
            display_sql_results((i.values() for i in ret), ["id"] + cmd["fields"])
            exit(1)

        if (
            cmd["cmd"] in ("drop_cmd", "import_cmd", "create_cmd", "set_cmd", "bed_cmd")
            or args.to_selection
        ):
            # PS: to_selection is used to detect select_cmd with selection creation
            display_query_status(ret)


if __name__ == "__main__":
    main()



import csv

from .abstractwriter import AbstractWriter
from cutevariant.core.sql import get_samples


class PedWriter(AbstractWriter):
    """Writer allowing to export samples of a project into a PED/PLINK file.

    Attributes:

        device: a file object typically returned by open("w")

    Example:
        >>> with open(filename,"rw") as file:
        ...    writer = MyWriter(file)
        ...    writer.save(conn)
    """

    def __init__(self, device):
        super().__init__(device)

    def save(self, conn, delimiter="\t", **kwargs):
        r"""Dump samples into a tabular file

        Notes:
            File is written without header.

        Example of line::

            `family_id\tindividual_id\tfather_id\tmother_id\tsex\tphenotype`

        Args:
            conn (sqlite.connection): sqlite connection
            delimiter (str, optional): Delimiter char used in exported file;
                (default: ``\t``).
            **kwargs (dict, optional): Arguments can be given to override
                individual formatting parameters in the current dialect.
        """
        writer = csv.DictWriter(
            self.device,
            delimiter=delimiter,
            lineterminator="\n",
            fieldnames=[
                "family_id",
                "name",
                "father_id",
                "mother_id",
                "sex",
                "phenotype",
            ],
            extrasaction="ignore",
            **kwargs
        )
        g = list(get_samples(conn))
        # Map DB ids with individual_ids
        individual_ids_mapping = {sample["id"]: sample["name"] for sample in g}
        # Add default value
        individual_ids_mapping[0] = 0
        # Replace DB ids
        for sample in g:
            sample["father_id"] = individual_ids_mapping[sample["father_id"]]
            sample["mother_id"] = individual_ids_mapping[sample["mother_id"]]
        writer.writerows(g)

    def save_from_list(self, samples, delimiter="\t", **kwargs):
        r"""Dump samples into a tabular file

        Args:
            samples(list): Iterable of samples; each sample is a list itself.
                => It's up to the user to give field in the correct order.
            delimiter (str, optional): Delimiter char used in exported file;
                (default: ``\t``).
            **kwargs (dict, optional): Arguments can be given to override
                individual formatting parameters in the current dialect.

        Notes:
            Replace None or empty strings to 0 (unknown PED ID)
        """
        writer = csv.writer(
            self.device, delimiter=delimiter, lineterminator="\n", **kwargs
        )
        # Replace None or empty strings to 0 (unknown PED ID)
        clean_samples = ([item if item else 0 for item in sample] for sample in samples)
        writer.writerows(clean_samples)



from PySide2.QtWidgets import QApplication

__title__ = "VQL Editor"
__description__ = "A VQL editor"
__long_description__ = QApplication.instance().translate(
    "vql_editor",
    """
<p>This plugin allows you to manage most of the work that can be done in a 
Cutevariant project by writing and executing VQL queries.</p>

<p>VQL is a <i>Domain Specific Language</i>, its main purpose is to filter variants
in the same fashion as a SQL query.</p>
<p>The VQL language can be run from the user interface via this plugin or directly
via the command line of Cutevariant.</p>

<p>Although the plugin offers auto-completion (do not forget the joker character '!')
and syntax highlighting, if you want to have more details on how to write your own VQL queries,
as well as the keywords of the language, please see the project wiki page:</p>

<a href="https://github.com/labsquare/cutevariant/wiki/VQL-language">
https://github.com/labsquare/cutevariant/wiki/VQL-language</a>
""",
)
__author__ = "Sacha schutz"
__version__ = "1.0.0"



import sys

from PySide2.QtWidgets import (
    QWidget,
    QSlider,
    QApplication,
    QStyleOptionSlider,
    QSizePolicy,
    QStyle,
)
from PySide2.QtCore import Qt, QRect, QSize, Signal
from PySide2.QtGui import QPainter, QPaintEvent, QMouseEvent, QPalette, QBrush


class RangeSlider(QWidget):

    rangeChanged = Signal(int, int)

    def __init__(self, parent=None):
        super().__init__(parent)

        self.first_position = 1
        self.second_position = 8

        self.tick_visible = False

        self.opt = QStyleOptionSlider()
        self.opt.minimum = 0
        self.opt.maximum = 10

        self.setTickPosition(QSlider.TicksAbove)
        self.setTickInterval(1)

        self.setSizePolicy(
            QSizePolicy(QSizePolicy.Expanding, QSizePolicy.Fixed, QSizePolicy.Slider)
        )

    def setRangeLimit(self, minimum: int, maximum: int):
        self.opt.minimum = minimum
        self.opt.maximum = maximum

    def setRange(self, start: int, end: int):
        self.first_position = start
        self.second_position = end

    def getRange(self):
        return (self.first_position, self.second_position)

    def setTickPosition(self, position: QSlider.TickPosition):
        self.opt.tickPosition = position

    def setTickInterval(self, ti: int):
        self.opt.tickInterval = ti

    def setTickVisible(self, visible: bool):
        self.tick_visible = visible

    def paintEvent(self, event: QPaintEvent):

        painter = QPainter(self)

        # Draw rule
        self.opt.initFrom(self)
        self.opt.rect = self.rect()
        self.opt.sliderPosition = 0

        if self.tick_visible:
            self.opt.subControls = QStyle.SC_SliderGroove | QStyle.SC_SliderTickmarks
        else:
            self.opt.subControls = QStyle.SC_SliderGroove

        #  Draw GROOVE
        self.style().drawComplexControl(QStyle.CC_Slider, self.opt, painter)

        # Draw INTERVAL

        color = self.palette().color(QPalette.Highlight)
        color.setAlpha(160)
        painter.setBrush(QBrush(color))
        painter.setPen(Qt.NoPen)

        self.opt.sliderPosition = self.first_position
        x_left_handle = (
            self.style()
            .subControlRect(QStyle.CC_Slider, self.opt, QStyle.SC_SliderHandle)
            .right()
        )

        self.opt.sliderPosition = self.second_position
        x_right_handle = (
            self.style()
            .subControlRect(QStyle.CC_Slider, self.opt, QStyle.SC_SliderHandle)
            .left()
        )

        groove_rect = self.style().subControlRect(
            QStyle.CC_Slider, self.opt, QStyle.SC_SliderGroove
        )

        selection = QRect(
            x_left_handle,
            groove_rect.y(),
            x_right_handle - x_left_handle,
            groove_rect.height(),
        ).adjusted(-1, 1, 1, -1)

        painter.drawRect(selection)

        # Draw first handle

        self.opt.subControls = QStyle.SC_SliderHandle
        self.opt.sliderPosition = self.first_position
        self.style().drawComplexControl(QStyle.CC_Slider, self.opt, painter)

        # Draw second handle
        self.opt.sliderPosition = self.second_position
        self.style().drawComplexControl(QStyle.CC_Slider, self.opt, painter)

    def mousePressEvent(self, event: QMouseEvent):

        self.opt.sliderPosition = self.first_position
        self._first_sc = self.style().hitTestComplexControl(
            QStyle.CC_Slider, self.opt, event.pos(), self
        )

        self.opt.sliderPosition = self.second_position
        self._second_sc = self.style().hitTestComplexControl(
            QStyle.CC_Slider, self.opt, event.pos(), self
        )

    def mouseMoveEvent(self, event: QMouseEvent):

        distance = self.opt.maximum - self.opt.minimum

        pos = self.style().sliderValueFromPosition(
            0, distance, event.pos().x(), self.rect().width()
        )

        if self._first_sc == QStyle.SC_SliderHandle:
            if pos <= self.second_position:
                self.first_position = pos
                self.update()
                return

        if self._second_sc == QStyle.SC_SliderHandle:
            if pos >= self.first_position:
                self.second_position = pos
                self.update()

    def mouseReleaseEvent(self, event: QMouseEvent):

        self.rangeChanged.emit(self.first_position, self.second_position)

    def sizeHint(self):
        """ override """
        SliderLength = 84
        TickSpace = 5

        w = SliderLength
        h = self.style().pixelMetric(QStyle.PM_SliderThickness, self.opt, self)

        if (
            self.opt.tickPosition & QSlider.TicksAbove
            or self.opt.tickPosition & QSlider.TicksBelow
        ):
            h += TickSpace

        return (
            self.style()
            .sizeFromContents(QStyle.CT_Slider, self.opt, QSize(w, h), self)
            .expandedTo(QApplication.globalStrut())
        )


if __name__ == "__main__":

    app = QApplication(sys.argv)

    w = RangeSlider()
    w.show()

    # q = QSlider()
    # q.show()

    app.exec_()



from PySide2.QtWidgets import *
from PySide2.QtCore import *
from PySide2.QtGui import *

import sys


class Pane(QFrame):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.widget = None
        self.label = QLabel()

        self.bar = QFrame()
        self.bar_layout = QHBoxLayout()
        self.bar_layout.setContentsMargins(0, 0, 0, 0)

        self.bar.setLayout(self.bar_layout)
        self.bar.setFrameShape(QFrame.NoFrame)

        # self.toolbar.setIconSize(QSize(16,16))
        # self.toolbar.setStyleSheet("background-color:palette(shadow); color: palette(light)")

        self.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Minimum)
        self.setFrameShape(QFrame.Panel)
        self.setFrameShadow(QFrame.Raised)
        self.setLineWidth(1)
        self.setMidLineWidth(3)

        # build expand buttons
        expand_button = QToolButton()
        expand_button.setAutoRaise(True)
        expand_button.setIcon(
            self.style().standardIcon(QStyle.SP_TitleBarUnshadeButton)
        )
        expand_button.setIconSize(QSize(14, 14))
        expand_button.clicked.connect(self.toggle_visible)

        # setup label
        # self.label.setText(self.windowTitle())
        font = QFont()
        font.setBold(True)
        self.label.setFont(font)
        self.bar_layout.addWidget(expand_button)

        self.bar_layout.addWidget(self.label)
        # self.spacer = QWidget()
        # self.spacer.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        # self.toolbar.addWidget(self.spacer)

        self.main_layout = QVBoxLayout()
        self.main_layout.addWidget(self.bar)
        self.main_layout.setContentsMargins(0, 0, 0, 0)
        self.main_layout.setSpacing(0)

        self.setLayout(self.main_layout)

    def toggle_visible(self):
        if self.widget:
            self.widget.setVisible(not self.widget.isVisible())

    def setWidget(self, widget):
        self.widget = widget
        self.main_layout.addWidget(self.widget)

    def setTitle(self, title: str):
        self.label.setText(title)


class TestPane(Pane):
    def create_widget(self):

        w = QWidget()
        l = QFormLayout()

        l.addRow("input", QSpinBox())
        l.addRow("frequenies", QSlider(Qt.Horizontal))
        w.setLayout(l)

        return w


class PanelListWidget(QScrollArea):
    def __init__(self, parent=None):
        super().__init__()
        self.panels = []
        self.viewport = QWidget()
        self.vlayout = QVBoxLayout()
        self.vlayout.setSpacing(1)
        self.vlayout.setContentsMargins(0, 0, 0, 0)
        self.viewport.setLayout(self.vlayout)
        self.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self.spacer = QWidget()
        self.spacer.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        self.vlayout.addWidget(self.spacer)
        self.vlayout.setSizeConstraint(QLayout.SetMaximumSize)

        self.setWidgetResizable(True)

        self.main_widget = QWidget()
        self.main_widget.setLayout(self.vlayout)

        self.setWidget(self.main_widget)

    def add(self, panel):
        self.panels.append(panel)
        self.vlayout.insertWidget(0, panel)
        panel.parent = self

    def clear(self):
        raise NotImplementedError("To Do @see #31")

    def indexOf(self, panel):
        self.panels.index(panel)


if __name__ == "__main__":

    app = QApplication(sys.argv)

    all_w = QTabWidget()

    w = PanelListWidget()

    w.add(TestPane())
    w.add(TestPane())
    w.add(TestPane())

    w2 = PanelListWidget()

    w2.add(TestPane())
    w2.add(TestPane())
    w2.add(TestPane())

    all_w.addTab(w, "Population")
    all_w.addTab(w2, "Samples")

    all_w.show()

    app.exec_()



"""Plugin to show all characteristics of a selected variant

VariantInfoWidget is showed on the GUI, it uses VariantPopupMenu to display a
contextual menu about the variant which is selected.
VariantPopupMenu is also used in viewquerywidget for the same purpose.
"""
from logging import DEBUG

# Qt imports
from PySide2.QtCore import Qt, Slot, QSize
from PySide2.QtWidgets import *
from PySide2.QtGui import QFont

# Custom imports
from cutevariant.gui import FIcon, style
from cutevariant.core import sql, get_sql_connection
from cutevariant.gui.plugin import PluginWidget
from cutevariant import commons as cm

from cutevariant.gui.widgets import DictWidget

LOGGER = cm.logger()


class EditPanel(QFrame):
    """Edit Panel 
    
    A panel box to edit a variant 

    get_data: returned updated variant with favorite, comment and classification 
    set_data: Fill the formular a variant  

    """

    # A signal emit when save button is pressed
    saved = Signal()

    _form_changed = Signal()

    def __init__(self, parent=None):
        super().__init__(parent)

        # Create fav button
        self.fav_button = QToolButton()
        self.fav_button.setCheckable(True)
        self.fav_button.setAutoRaise(True)
        self.fav_button.clicked.connect(self._form_changed)
        icon = QIcon()
        icon.addPixmap(FIcon(0xF00C3).pixmap(32, 32), QIcon.Normal, QIcon.Off)
        icon.addPixmap(FIcon(0xF00C0).pixmap(32, 32), QIcon.Normal, QIcon.On)
        self.fav_button.setIcon(icon)

        # Create classification combobox
        self.class_edit = QComboBox()
        self.class_edit.setFrame(False)
        self.class_edit.currentIndexChanged.connect(self._form_changed)

        for key in style.CLASSIFICATION:
            self.class_edit.addItem(
                FIcon(
                    style.CLASSIFICATION[key]["icon"],
                    style.CLASSIFICATION[key]["color"],
                ),
                style.CLASSIFICATION[key]["name"],
            )

        # Create comment form . This is a stack widget with a PlainText editor
        # and a Markdown preview as QTextBrowser

        self.comment_edit = QPlainTextEdit()
        self.comment_edit.setPlaceholderText("Write a comment in markdown ...")
        self.comment_edit.textChanged.connect(self._form_changed)

        self.comment_preview = QTextBrowser()
        self.comment_preview.setPlaceholderText("Press edit to add a comment ...")
        self.comment_preview.setAlignment(Qt.AlignTop | Qt.AlignLeft)
        self.comment_preview.setOpenExternalLinks(True)
        self.comment_preview.setFrameStyle(QFrame.NoFrame)

        # Create stacked
        self.stack = QStackedWidget()
        self.stack.addWidget(self.comment_edit)
        self.stack.addWidget(self.comment_preview)
        self.stack.setCurrentIndex(1)

        # Build layout
        self.setFrameShape(QFrame.StyledPanel)

        # header  layout
        title_layout = QHBoxLayout()
        title_layout.addWidget(self.class_edit)
        title_layout.addWidget(self.fav_button)

        # Button layout
        self.switch_button = QPushButton(self.tr("Edit comment..."))
        self.switch_button.setFlat(True)
        self.save_button = QPushButton(self.tr("Save"))
        self.save_button.setFlat(True)
        bar_layout = QHBoxLayout()
        bar_layout.addWidget(self.switch_button)
        bar_layout.addStretch()
        bar_layout.addWidget(self.save_button)

        # Main layout
        v_layout = QVBoxLayout()
        v_layout.addLayout(title_layout)
        v_layout.addWidget(self.stack)
        v_layout.addLayout(bar_layout)
        self.setLayout(v_layout)

        # Create connection
        self.switch_button.clicked.connect(self.switch_mode)
        self.save_button.clicked.connect(self._on_save)
        self._form_changed.connect(lambda: self.save_button.setEnabled(True))

    def switch_mode(self):
        """Switch comment mode between editable and previewer 
        """
        if self.stack.currentWidget() == self.comment_edit:
            self.stack.setCurrentIndex(1)
            self.comment_preview.setMarkdown(str(self.comment_edit.toPlainText()))
            self.switch_button.setText(self.tr("Edit comment..."))
        else:
            self.stack.setCurrentIndex(0)
            self.switch_button.setText(self.tr("Preview comment..."))

    def _on_save(self):
        """ private slot reacting when save_button is pressed """

        if self.stack.currentWidget() == self.comment_edit:
            self.switch_mode()
        self.saved.emit()
        self.save_button.setEnabled(False)

    def set_text(self, text: str):
        """Set comment 
        
        Args:
            text (str): Description
        
        """
        if text is None:
            self.clear()
            return

        self.comment_edit.setPlainText(str(text))
        try:
            self.comment_preview.setMarkdown(str(text))
        except AttributeError:
            # Fallback Qt 5.14-
            self.comment_preview.setPlainText(str(text))

    def clear(self):
        """Clear comment 
        """
        self.comment_edit.clear()
        self.comment_preview.clear()

    def set_data(self, variant: dict):
        """Set Form with variant data

        Args:
            variant (dict): 
            
        Exemples:            
            w.set_data({"favorite": 1, "comment":'salut', "classification":4})
        """

        self.clear()
        if "classification" in variant:
            self.class_edit.setCurrentIndex(variant["classification"])

        if "comment" in variant:
            self.set_text(variant["comment"])

        if "favorite" in variant:
            self.fav_button.setChecked(
                Qt.Checked if variant["favorite"] else Qt.Unchecked
            )

        self.save_button.setEnabled(False)

    def get_data(self) -> dict:
        """Get variant data from Form input 
        
        Returns:
            dict: variant updated data
        """
        return {
            "classification": self.class_edit.currentIndex(),
            "comment": self.comment_edit.toPlainText(),
            "favorite": int(self.fav_button.isChecked()),
        }


class VariantInfoWidget(PluginWidget):
    """Plugin to show all annotations of a selected variant"""

    ENABLE = True

    def __init__(self, parent=None):
        super().__init__(parent)

        self.setWindowTitle(self.tr("Info variants"))

        # Current variant => set by on_refresh and on_open_project
        self.current_variant = None

        self.view = QTabWidget()
        self.toolbar = QToolBar()
        self.toolbar.setIconSize(QSize(16, 16))

        # Build comments tab
        self.edit_panel = EditPanel()
        self.edit_panel.saved.connect(self.on_save_variant)
        self.view.addTab(self.edit_panel, self.tr("User"))

        # Build variant tab
        self.variant_view = DictWidget()
        self.variant_view.set_header_visible(True)

        self.view.addTab(self.variant_view, self.tr("Variant"))

        # Build transcript tab
        self.transcript_combo = QComboBox()
        self.transcript_view = DictWidget()
        self.transcript_view.set_header_visible(True)
        tx_layout = QVBoxLayout()
        tx_layout.addWidget(self.transcript_combo)
        tx_layout.addWidget(self.transcript_view)
        tx_widget = QWidget()
        tx_widget.setLayout(tx_layout)
        self.view.addTab(tx_widget, self.tr("Transcripts"))
        self.transcript_combo.currentIndexChanged.connect(self.on_transcript_changed)

        # Build Samples tab
        self.sample_combo = QComboBox()
        self.sample_view = DictWidget()
        self.sample_view.set_header_visible(True)

        tx_layout = QVBoxLayout()
        tx_layout.addWidget(self.sample_combo)
        tx_layout.addWidget(self.sample_view)
        tx_widget = QWidget()
        tx_widget.setLayout(tx_layout)
        self.view.addTab(tx_widget, self.tr("Samples"))
        self.sample_combo.currentIndexChanged.connect(self.on_sample_changed)

        # Build genotype tab
        self.genotype_view = QListWidget()
        self.genotype_view.setIconSize(QSize(20, 20))
        self.view.addTab(self.genotype_view, self.tr("Genotypes"))

        v_layout = QVBoxLayout()
        v_layout.setContentsMargins(0, 0, 0, 0)
        v_layout.addWidget(self.view)
        self.setLayout(v_layout)

        # # Create menu
        # TODO: restore this
        # self.context_menu = VariantPopupMenu()
        # # Ability to trigger the menu
        # self.view.setContextMenuPolicy(Qt.CustomContextMenu)
        # self.view.customContextMenuRequested.connect(self.show_menu)
        # self.add_tab("variants")

        # Cache all database fields and their descriptions for tooltips
        # Field names as keys, descriptions as values
        self.fields_descriptions = None

        # Cache genotype icons
        # Values in gt field as keys (str), FIcon as values
        self.genotype_icons = {
            key: FIcon(val) for key, val in cm.GENOTYPE_ICONS.items()
        }

    def clear(self):
        """ Clear all view """
        self.variant_view.clear()
        self.transcript_view.clear()
        self.sample_view.clear()
        self.genotype_view.clear()
        self.edit_panel.clear()

    def on_save_variant(self):

        # if view is visible
        if "variant_view" in self.mainwindow.plugins:
            variant_view = self.mainwindow.plugins["variant_view"]

            update_data = self.edit_panel.get_data()

            index = variant_view.main_right_pane.view.currentIndex()
            variant_view.main_right_pane.model.update_variant(index.row(), update_data)

            # variant_view.main_right_pane.model.update_variant(index.row(), update_data)

        else:
            # TODO BUT UNNECESSARY because we always have a variant_viex ...
            # Save directly in database ?
            pass

    def on_open_project(self, conn):
        self.conn = conn
        self.clear()
        self.on_refresh()
        # Cache DB fields descriptions
        self.fields_descriptions = {
            field["name"]: field["description"] for field in sql.get_fields(self.conn)
        }

    def on_refresh(self):
        """Set the current variant by the variant displayed in the GUI"""
        self.current_variant = self.mainwindow.state.current_variant
        self.populate()

    def populate(self):
        """Show the current variant attributes on the TreeWidget

        .. note:: RichText in Markdown is supported starting PySide 5.14
        """
        if not self.current_variant or "id" not in self.current_variant:
            return

        variant_id = self.current_variant["id"]

        # Populate variant
        data = dict(
            [
                (k, v)
                for k, v in sql.get_one_variant(self.conn, variant_id).items()
                if k not in ("variant_id", "sample_id", "annotations", "samples")
            ]
        )
        self.variant_view.set_dict(data)

        self.edit_panel.set_data(data)

        title = "{chr}:{pos} {ref}>{alt}".format(**data)
        # self.parent().setWindowTitle(title)

        # Populate annotations
        self.transcript_combo.blockSignals(True)
        self.transcript_combo.clear()

        for annotation in sql.get_annotations(self.conn, variant_id):
            if "transcript" in annotation:
                self.transcript_combo.addItem(annotation["transcript"], annotation)
        self.on_transcript_changed()
        self.transcript_combo.blockSignals(False)

        # Populate samples
        self.sample_combo.blockSignals(True)
        self.sample_combo.clear()
        for sample in sql.get_samples(self.conn):
            self.sample_combo.addItem(sample["name"], sample["id"])
        self.on_sample_changed()
        self.sample_combo.blockSignals(False)

        # Populate genotypes for samples
        self.genotype_view.clear()
        query = f"""SELECT samples.name, sv.gt FROM samples 
                 LEFT JOIN sample_has_variant sv ON samples.id = sv.sample_id 
                 AND sv.variant_id = {variant_id}"""

        for sample_name, genotype in self.conn.execute(query):
            item = QListWidgetItem()
            icon = self.genotype_icons.get(genotype, self.genotype_icons[-1])

            item.setText(sample_name)
            item.setIcon(icon)
            item.setToolTip(cm.GENOTYPE_DESC.get(genotype, -1))

            self.genotype_view.addItem(item)

    @Slot()
    def on_transcript_changed(self):
        """This method is triggered when transcript change from combobox

        Fill fields & values for selected transcript.
        """
        annotations = self.transcript_combo.currentData()
        self.transcript_view.set_dict(annotations)

    @Slot()
    def on_sample_changed(self):
        """This method is triggered when sample change from combobox

        Fill fields & values for selected sample.
        """
        sample_id = self.sample_combo.currentData()
        variant_id = self.current_variant["id"]

        self.sample_view.set_dict(
            sql.get_sample_annotations(self.conn, variant_id, sample_id)
        )

    # @Slot()
    # def on_save_clicked(self):
    #     """Save button
    #     """
    #     classification = self.classification_box.currentIndex()
    #     favorite = self.favorite_checkbox.isChecked()
    #     comment = self.comment_input.toPlainText()

    #     updated = {
    #         "classification": classification,
    #         "favorite": favorite,
    #         "comment": comment,
    #     }

    #     if "variant_view" in self.mainwindow.plugins:
    #         main_view = self.mainwindow.plugins["variant_view"]
    #         print(main_view)

    # sql.update_variant(self.conn, updated)

    # def show_menu(self, pos: QPoint):
    #     """Show context menu associated to the current variant"""
    #     if not self.current_variant:
    #         return
    #     self.context_menu.popup(self.current_variant, self.view.mapToGlobal(pos))


if __name__ == "__main__":
    import sys

    app = QApplication(sys.argv)

    conn = get_sql_connection("/home/schutz/Dev/cutevariant/examples/test.db")

    w = VariantInfoWidget()
    w.conn = conn

    variant = sql.get_one_variant(conn, 1)

    w.current_variant = variant

    w.show()

    app.exec_()



from PySide2.QtWidgets import (
    QWidget,
    QSlider,
    QApplication,
    QHBoxLayout,
    QAbstractSpinBox,
    QSpinBox,
)
from PySide2.QtCore import Qt

import sys


class EditSlider(QWidget):
    def __init__(self, parent=None):
        super().__init__(parent)

        hlayout = QHBoxLayout()
        hlayout.setContentsMargins(0, 0, 0, 0)
        hlayout.setSpacing(2)

        self.edit = QSpinBox()
        self.edit.setMinimumWidth(40)
        self.edit.setButtonSymbols(QAbstractSpinBox.NoButtons)
        self.slider = QSlider(Qt.Horizontal)

        hlayout.addWidget(self.edit)
        hlayout.addWidget(self.slider)

        self.setLayout(hlayout)

        self.slider.valueChanged.connect(self.setValue)
        self.edit.valueChanged.connect(self.setValue)

    def setRange(self, min_val, max_val):
        self.edit.setRange(min_val, max_val)
        self.slider.setRange(min_val, max_val)

    def value(self) -> int:
        return self.edit.value()

    def setValue(self, value: int):

        if self.sender() != self.slider and self.sender() != self.edit:
            self.edit.setValue(value)
            self.slider.setValue(value)
            return

        if self.sender() == self.slider:
            self.edit.setValue(value)

        if self.sender() == self.edit:
            self.slider.setValue(value)


if __name__ == "__main__":

    app = QApplication(sys.argv)

    w = EditSlider()
    w.show()

    # q = QSlider()
    # q.show()

    app.exec_()



from PySide2.QtWidgets import QApplication

__title__ = "WordSets"
__description__ = "A plugin to manage word sets"
__long_description__ = QApplication.instance().translate(
    "word_sets",
    """
<p>This plugin allows to create sets of words that can be matched with the
attributes of the project's variants.</p>
<p>
Once the addition of a word set is started, a manual addition one by one of the
words is possible; for practical reasons it is however advisable to directly
import a text file containing merely 1 word per line.</p>

The set can be reworked at any time via an editor.<br>
<br>
<i>Example of use:</i><br>
<br>
<i>A user wishes to quickly filter all variants of a project related to a set of
relevant genes for him.
He therefore creates a word set and then makes a selection via:</i>

<ul>
<li>the <em>Filters Editor</em> plugin with a filter of the type:
<pre>gene IN ('WORDSET', 'my_word_set')</pre></li>

<li>the <em>VQL Editor</em> plugin with a VQL request of the type:
<pre>SELECT chr,pos,ref,alt,gene FROM variants WHERE gene IN WORDSET['my_word_set']</pre></li>
</ul>
""",
)
__author__ = "Sacha schutz"
__version__ = "1.0.0"



from abc import ABC, abstractmethod
import io
import gzip
import math
from collections import Counter
import cutevariant.commons as cm

LOGGER = cm.logger()


class AbstractReader(ABC):
    """Base class for all Readers required to import variants into the database.

    Subclass it if you want a new file parser .

    Attributes:

        device: A file object typically returned by open(); Can be None if
            FakeReader type is instanciated.
        file_size: File size in bytes
            See Also: :meth:`self.get_total_file_size`
        number_lines: Number of lines in the file (compressed or not).
            See Also: :meth:`self.compute_number_lines`
        read_bytes: Current bytes readed (progression = read_bytes / file_size)
            It's a fallback if number_lines can't be computed.
        samples: List of samples in the file (default: empty)

        ignored_fields: Skip fields in importations. 
            A list of fields to skip [{field_name:"AF", "category":"variant"}]

    Example:
        >>> with open(filename,"r") as file:
        ...    reader = Reader(file)
        ...    reader.get_variants()
    """

    def __init__(self, device):
        self.device = device
        self.number_lines = None
        self.read_bytes = 0
        self.samples = list()

        self.file_size = self.get_total_file_size()
        self.compute_number_lines()

        self.ignored_fields = set()

    @classmethod
    @abstractmethod
    def get_variants(cls):
        """Abstract method must return variants as an iterable of dictionnaries.

        Variant dictionnary has 4 mandatory fields `chr`, `pos`, `ref`, `alt`.
        Other fields are optionnal
        For instance::

            [
                {"chr": "chr3","pos": 3244,"ref": "A","alt":"C", "qual": 30},
                {"chr": "chr4","pos": 3244,"ref": "A","alt":"C","qual": 20},
                {"chr": "chr5","pos": 3244,"ref": "A","alt":"C","qual": 10 },
            ]

        Annotations and Samples objects can be embbeded into a variant dictionnaries.
        Annotations describes several annotations for one variant.
        In the most of the case, those are relative to transcripts.
        Samples describes information relative to a variant with a sample,
        like genotype (gt). This is a mandatory field.

        .. code-block:: python

            [{
                "chr": "chr3",
                "pos": 3244,
                "ref": "A",
                "alt":"C",
                "field_n": "value_n",
                "annotations": [
                    {"gene": "GJB2", "transcripts": "NM_00232.1", "field_n": "value_n"},
                    {"gene": "GJB2", "transcripts": "NM_00232.2", "field_n": "value_n"}
                ],
                "samples": [
                    {"name":"boby", "genotype": 1, "field_n":"value_n"},
                    {"name":"kevin", "genotype": 1, "field_n":"value_n"}
                ]
            },]

        Yields:
            dict: variant dictionnary

        Examples:
            >>> for variant in reader.get_variants():
            ...     print(variant["chr"], variant["pos"])

        """
        raise NotImplementedError(cls.__class__.__name__)

    @classmethod
    @abstractmethod
    def get_fields(cls):
        """Abstract method hat must return fields description

        Full output::

            [
            {"name": "chr", "type": "text", "category": "variant", "description": "..."},
            {"name": "pos", "type": "text", "category": "variant", "description": "..."},
            {"name": "ref", "type": "text", "category": "variant", "description": "..."},
            {"name": "alt", "type": "text", "category": "variant", "description": "..."},
            {"name": "field_n", "type": "text", "category": "variant", "description": "..."},
            {"name": "name", "type": "text", "category": "annotations", "samples": "..."},
            {"name": "genotype", "type": "text", "category": "annotations", "samples": "..."}
            ]

        Yields:
            dict: field dictionnary

        Examples:
            >>> for field in reader.get_fields():
            ...     print(field["name"], field["description"])
        """
        raise NotImplementedError(cls.__class__.__name__)

    def get_samples(self) -> list:
        """Return list of samples

        Override this method to have samples in sqlite database.
        """
        return []

    def get_metadatas(self) -> dict:
        """Get meta data

        Override this method to have meta data in sqlite database
        """
        return {}

    def get_extra_fields(self):
        """Yield fields with extra mandatory fields like 'comment' and 'score'"""
        yield {
            "name": "favorite",
            "type": "bool",
            "category": "variants",
            "description": "Tag status as favorite",
        }
        yield {
            "name": "comment",
            "type": "str",
            "category": "variants",
            "description": "Variant comment written by user",
        }
        yield {
            "name": "classification",
            "type": "int",
            "category": "variants",
            "description": "ACMG score",
        }

        yield {
            "name": "count_hom",
            "type": "int",
            "category": "variants",
            "description": "Number of homozygous genotypes (1/1)",
        }
        yield {
            "name": "count_het",
            "type": "int",
            "category": "variants",
            "description": "Number of heterozygous genotypes (0/1)",
        }
        yield {
            "name": "count_ref",
            "type": "int",
            "category": "variants",
            "description": "Number of homozygous genotypes (0/0)",
        }
        yield {
            "name": "count_var",
            "type": "int",
            "category": "variants",
            "description": "Number of variants (not 0/0)",
        }

        yield {
            "name": "is_indel",
            "type": "bool",
            "category": "variants",
            "description": "True if variant is an indel",
        }

        yield {
            "name": "is_snp",
            "type": "bool",
            "category": "variants",
            "description": "True if variant is a snp",
        }

        yield {
            "name": "annotation_count",
            "type": "int",
            "category": "variants",
            "description": "Count of transcripts per variant",
        }

        yield {
            "name": "case_count_hom",
            "type": "int",
            "category": "variants",
            "description": "Number of homozygous genotypes (1/1) in case",
        }

        yield {
            "name": "case_count_het",
            "type": "int",
            "category": "variants",
            "description": "Number of heterozygous genotypes (1/0) in case",
        }

        yield {
            "name": "case_count_ref",
            "type": "int",
            "category": "variants",
            "description": "Number of homozygous genotypes (0/0) in case",
        }

        yield {
            "name": "control_count_hom",
            "type": "int",
            "category": "variants",
            "description": "Number of homozygous genotypes (1/1) in control",
        }

        yield {
            "name": "control_count_het",
            "type": "int",
            "category": "variants",
            "description": "Number of heterozygous genotypes (1/0) in control",
        }

        yield {
            "name": "control_count_ref",
            "type": "int",
            "category": "variants",
            "description": "Number of homozygous genotypes (0/0) in control",
        }

        # avoid duplicates fields by categories ...
        duplicates = set()
        for field in self.get_fields():
            # Create unique identifiant by categories
            unique_key = field["category"] + "." + field["name"]
            is_unique = unique_key not in duplicates
            is_ignored = (field["name"], field["category"]) in self.ignored_fields
            if is_unique and not is_ignored:
                yield field

            duplicates.add(unique_key)

    def add_ignored_field(self, field_name: str, field_category: str):
        """Add new field to the ignored_fields list. 
        ignored fields will not returned by get_extra_fields and then are not imporpted 
        into the database 
        
        Args:
            field_name (str): a field name 
            field_category (str): the category field name (variant,annotation,sample)
        """

        self.ignored_fields.add((field_name, field_category))

    def get_extra_variants(self, **kwargs):
        """Yield variants with extra information computed.

        The following information are added. See get_extra_fields

        - favorite (bool): (Default: False)
        - comment (str): (Default: "")
        - classification (int): ACMG score (Default: 3 (uncertain significance)
        - count_var (int): Number of variants (not 0/0)
        - count_hom (int): How many variants are mutant homozygous within samples
        - count_het (int): How many variants are heterozygous within samples
        - count_ref (int): How many variants are wild homozygous with samples
        - is_indel (bool): Is the variation an insertion / deletion
        - is_snp (bool): Is the variation an single nucleotide variation

        If case/control are available from a pedfile, counting from case and
        control is also computed.
        In this case, it is necessary to give sample names in "case" and
        "control" keys in kwargs .

        Example of supported kwargs::

            {
                "case": ["boby", "raymond"],
                "control": ["lucas", "pierre"]
            }

        - case_count_hom (int): How many variants are mutant homozygous within case samples
        - case_count_het (int): How many variants are heterozygous within case samples
        - case_count_ref (int): How many variants are wild heterozygous within case samples
        - control_count_hom (int): How many variants are mutant homozygous within control samples
        - control_count_het (int): How many variants are heterozygous within control samples
        - control_count_ref (int): How many variants are wild heterozygous within control samples

        See Also:
            `cutevariant.core.reader.vcfreader.VcfReader.parse_variants`

        Args:
            **kwargs (optional): case and control sample names

        Yields:
            (generator[dict]): variants. See also: :meth:`get_variants`.

        Raises:
            AssertionError: If sample(s) are both in cases and controls.
        """
        case_and_control_samples_found = False
        if "case" in kwargs and "control" in kwargs:
            # Samples can't be both in cases and controls
            case_samples = kwargs["case"]
            control_samples = kwargs["control"]
            assert not set(case_samples) & set(
                control_samples
            ), "Found sample both in cases and controls!"
            case_and_control_samples_found = True

        for variant in self.get_variants():
            variant["favorite"] = False
            variant["comment"] = ""
            variant["classification"] = 3

            # For now set the first annotation as a major transcripts
            if "annotations" in variant:
                variant["annotation_count"] = len(variant["annotations"])

            # Count genotype by control and case
            genotype_counter = Counter()
            if "samples" in variant:
                for sample in variant["samples"]:
                    genotype_counter[sample["gt"]] += 1

            variant["count_hom"] = genotype_counter[2]
            variant["count_het"] = genotype_counter[1]
            variant["count_ref"] = genotype_counter[0]
            # Number of variants (not 0/0)
            variant["count_var"] = genotype_counter[1] + genotype_counter[2]

            variant["is_indel"] = len(variant["ref"]) != len(variant["alt"])
            variant["is_snp"] = len(variant["ref"]) == len(variant["alt"])

            # Count genotype by control and case
            if case_and_control_samples_found:

                case_counter = Counter()
                control_counter = Counter()

                if "samples" in variant:
                    # Note: No garantee that samples from DB are all qualified
                    # by PED data.
                    # So some samples from variants may not be in case/control samples.
                    for sample in variant["samples"]:
                        if sample["name"] in case_samples:
                            case_counter[sample["gt"]] += 1

                        elif sample["name"] in control_samples:
                            control_counter[sample["gt"]] += 1

                variant["case_count_hom"] = case_counter[2]
                variant["case_count_het"] = case_counter[1]
                variant["case_count_ref"] = case_counter[0]

                variant["control_count_hom"] = control_counter[2]
                variant["control_count_het"] = control_counter[1]
                variant["control_count_ref"] = control_counter[0]

            # Remove ignore variants
            for name, category in self.ignored_fields:
                if category == "variants":
                    if name in variant:
                        del variant[name]
                # remove from category
                if category == "annotations":
                    for ann in variant["annotations"]:
                        if name in ann:
                            del ann[name]

                if category == "samples":
                    for sample in variant["samples"]:
                        if name in sample:
                            del sample[name]

            yield nullify(variant)

    def get_extra_fields_by_category(self, category: str):
        """Syntaxic suggar to get fields according their category

        :param category can be usually variants, samples, annotations
        :return: A generator of fields
        :rtype: <generator>
        """
        return (
            field for field in self.get_extra_fields() if field["category"] == category
        )

    def get_fields_by_category(self, category: str):
        """Syntaxic suggar to get fields according their category

        :param category can be usually variants, samples, annotations
        :return: A generator of fields
        :rtype: <generator>
        """
        return (field for field in self.get_fields() if field["category"] == category)

    def get_variants_count(self) -> int:
        """Get variant count from the device

        Override this method to make it faster
        """
        return len(tuple(self.get_variants()))

    def get_total_file_size(self) -> int:
        """Compute file size int bytes"""
        # FakeReader is used ?
        if not self.device:
            return 0

        filename = self.device.name

        if cm.is_gz_file(filename):
            return cm.get_uncompressed_size(filename)
        # Go to EOF and get position in bytes
        size = self.device.seek(0, 2)
        # Rewind the file
        self.device.seek(0)
        return size

    def compute_number_lines(self):
        """Get a sample of lines in file if possible and if the end of file is
        not reached compute an evaluation of the global number of lines.

        Returns:
            Nothing but sets `self.number_lines` attribute.
        """

        def find_lines_in_text_file(text_file_handler):
            """Get first 15000 lines

            PS: don't care of headers (# lines), the influence is marginal on big
            files and also on small files (so quick to insert that the wrong number
            of lines is invisible).
            """
            first_lines = []
            for _ in range(15000):
                try:
                    first_lines.append(len(next(text_file_handler)))
                except StopIteration:
                    # EOF: exact number of lines is known
                    self.number_lines = len(first_lines)
                    break

            if self.number_lines is None:
                self.number_lines = int(
                    self.file_size / (sum(first_lines) / len(first_lines))
                )

            LOGGER.debug(
                "nb lines evaluated: %s; size: %s; lines used: %s",
                self.number_lines,
                self.file_size,
                len(first_lines),
            )

        # FakeReader is used ?
        if not self.device:
            return 0

        # Detect type of file handler
        if isinstance(self.device, (io.RawIOBase, io.BufferedIOBase)):
            # Binary opened file => assert that it is a vcf.gz file
            with gzip.open(self.device.name, "rb") as file_obj:
                find_lines_in_text_file(file_obj)
        elif isinstance(self.device, io.TextIOBase):
            find_lines_in_text_file(self.device)
        else:
            LOGGER.error("Unknown file handler type: %s", type(self.device))
            raise TypeError("Unknown file handler type: %s" % type(self.device))

        # Rewind the file
        self.device.seek(0)


def check_variant_schema(variant: dict):
    """Test if get_variant returns well formated nested data.

    This method is for testing purpose. It raises an exception if data is corrupted

    :param variant dict returned by AbstractReader.get_variant()

    """
    try:
        from schema import Schema, And, Or, Use, Optional, Regex
    except ImportError as e:
        LOGGER.warning("You should install optional package 'schema' via:")
        LOGGER.warning("\t - pypi: pip install cutevariant[dev]")
        LOGGER.warning("\t - git repo in editable mode: pip -e . [dev]")
        raise e

    checker = Schema(
        {
            "chr": And(Use(str.lower), str),
            "pos": int,
            "ref": And(Use(str.upper), Regex(r"^[ACGTN]+")),
            "alt": And(Use(str.upper), Regex(r"^[ACGTN]+")),
            Optional(str): Or(int, str, bool, float, None),
            Optional("annotations"): [
                {
                    "gene": str,
                    "transcript": str,
                    Optional(str): Or(int, str, bool, float),
                }
            ],
            Optional("samples"): [
                {
                    "name": str,
                    "gt": And(int, lambda x: x in [-1, 0, 1, 2]),
                    Optional(str): Or(int, str, bool, float),
                }
            ],
        }
    )

    checker.validate(variant)


def check_field_schema(field: dict):
    """Test if get_field returns well formated data

    This method is for testing purpose. It raises an exception if data is corrupted

    :param field dict returned by AbstractReader.get_field()
    """
    try:
        from schema import Schema, And, Use, Optional
    except ImportError as e:
        LOGGER.warning("You should install optional package 'schema' via:")
        LOGGER.warning("\t - pypi: pip install cutevariant[dev]")
        LOGGER.warning("\t - git repo in editable mode: pip -e . [dev]")
        raise e

    checker = Schema(
        {
            "name": And(str, Use(str.lower)),
            "type": lambda x: x in ["str", "int", "bool", "float"],
            "category": lambda x: x in ["variants", "annotations", "samples"],
            "description": str,
            Optional("constraint", default="NULL"): str,
        }
    )

    checker.validate(field)


def sanitize_field_name(field: str):
    # TODO
    LOGGER.warning("NOT implemented function!!")
    return field


def nullify(variant: dict) -> dict:
    """ Convert empty fields value  to NONE 
    This is used have NULL value inside the SQLITE inside an empty string

    """

    def convert_to_none(value):
        """ convert value to None according type """
        EMPTY_STRING = ["", "."]
        if isinstance(value, str):
            if value in EMPTY_STRING:
                return None

        if isinstance(value, float) or isinstance(value, int):
            if math.isnan(value):
                return None

        return value

    for key in variant.keys():
        variant[key] = convert_to_none(variant[key])

        if key == "annotations":
            for ann in variant["annotations"]:
                for ann_key in ann.keys():
                    ann[ann_key] = convert_to_none(ann[ann_key])

        if key == "samples":
            for sample in variant["samples"]:
                for sample_key in sample.keys():
                    sample[sample_key] = convert_to_none(sample[sample_key])

    return variant



"""Plugin to view/edit/remove/do set operations on selections in the database
from the GUI.
SourceEditorWidget class is seen by the user and uses selectionModel class
as a model that handles records from the database.
"""
# Qt imports
from PySide2.QtWidgets import *
from PySide2.QtCore import *
from PySide2.QtGui import *

# Custom imports
from cutevariant.core import sql, command
from cutevariant.core.reader import BedReader
from cutevariant.gui import plugin, FIcon
from cutevariant.commons import logger, DEFAULT_SELECTION_NAME


LOGGER = logger()

# =================== SELECTION MODEL ===========================
class SourceModel(QAbstractTableModel):
    """Model to store all selections from SQLite `selections` table.
    Usage:
        model = selectionModel()
        model.conn = conn
        model.load()
    """

    def __init__(self, conn=None):
        super().__init__()
        self.conn = conn
        self.records = []

    def rowCount(self, parent=QModelIndex()):
        """Overrided from QAbstractTableModel"""
        return len(self.records)

    def columnCount(self, parent=QModelIndex()):
        """Overrided from QAbstractTableModel"""
        return 2  # value and count

    def data(self, index: QModelIndex(), role=Qt.DisplayRole):
        """Return the data stored under the given role for the item referred
        to by the index (row, col)
        Overrided from QAbstractTableModel
        :param index:
        :param role:
        :type index:
        :type role:
        :return: None if no valid index
        :rtype:
        """

        if not index.isValid():
            return None

        if role == Qt.DisplayRole:
            if index.column() == 0:
                return self.records[index.row()]["name"]

            if index.column() == 1:
                return self.records[index.row()]["count"]

        if role == Qt.DecorationRole:
            if index.column() == 0:
                return QIcon(FIcon(0xF04F1))

        return None

    def headerData(self, section, orientation, role=Qt.DisplayRole):
        """Return the data for the given role and section in the header with
        the specified orientation
        Overrided from QAbstractTableModel
        """
        if not self.records:
            return

        if orientation == Qt.Horizontal and role == Qt.DisplayRole:
            if section == 0:
                return "selection"
            if section == 1:
                return "count"

        if orientation == Qt.Vertical and role == Qt.DisplayRole and section > 0:
            return self.records[section].get(
                "id", None
            )  # For debug purpose . displayed in vertical header

    def record(self, index: QModelIndex()):
        """Return Selection records by index"""
        if not index.isValid():
            return None
        return self.records[index.row()]

    def find_record(self, name: str):
        """Find a record by name
        @see: view.selectionModel()
        :return: QModelIndex
        """
        for idx, record in enumerate(self.records):
            if record["name"] == name:
                return self.index(idx, 0, QModelIndex())
        return QModelIndex()

    def remove_record(self, index: QModelIndex()):
        """Delete the selection with the given id in the database
        :return: True if the deletion has been made, False otherwise.
        :rtype: <boolean>
        """
        # Get selected record
        record = self.record(index)

        self.beginRemoveRows(QModelIndex(), index.row(), index.row())

        if sql.delete_selection(self.conn, record["id"]):
            # Delete in model; triggers currentRowChanged signal
            #  Magic... the record disapear ...  ??
            self.endRemoveRows()
            return True
        return False

    def edit_record(self, index, record: dict):
        """Edit the given selection in the database and emit `dataChanged` signal"""
        if sql.edit_selection(self.conn, record):
            self.dataChanged.emit(index, index)

    def load(self):
        """Load all selections into the model"""

        if self.conn is None:
            return

        self.beginResetModel()
        # Add all selections from the database
        # Dictionnary of all attributes of the table.
        #    :Example: {"name": ..., "count": ..., "query": ...}
        self.records = list(sql.get_selections(self.conn))
        self.endResetModel()


# =================== SELECTION VIEW ===========================


class SourceEditorWidget(plugin.PluginWidget):
    """Widget displaying the list of avaible selections.
    User can select one of them to update Query::selection
    """

    ENABLE = True

    def __init__(self, parent=None, conn=None):
        """
        Args:
            parent (QMainWindow): Mainwindow of Cutevariant, passed during
                plugin initialization.
            conn (sqlite3.connexion): Sqlite3 connexion
        """
        super().__init__(parent)

        self.setWindowTitle(self.tr("Source editor"))
        # conn is always None here but initialized in on_open_project()
        self.model = SourceModel(conn)
        self.view = QTableView()
        self.view.setModel(self.model)
        self.view.setSelectionBehavior(QAbstractItemView.SelectRows)
        self.view.setSelectionMode(QAbstractItemView.SingleSelection)
        self.view.horizontalHeader().show()
        self.view.horizontalHeader().setStretchLastSection(False)
        self.view.horizontalHeader().setSectionResizeMode(
            0, QHeaderView.ResizeToContents
        )
        self.view.horizontalHeader().setSectionResizeMode(1, QHeaderView.Stretch)
        self.view.horizontalHeader().hide()

        self.toolbar = QToolBar()
        self.toolbar.setIconSize(QSize(16, 16))

        self.view.verticalHeader().hide()
        self.view.verticalHeader().setDefaultSectionSize(26)
        self.view.setShowGrid(False)
        self.view.setAlternatingRowColors(True)

        layout = QVBoxLayout()
        layout.addWidget(self.toolbar)
        layout.addWidget(self.view)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(0)

        self.setLayout(layout)

        # Used to block signals during the insertions (if set to True)
        self.is_loading = False

        # Map the operations of context menu with an internal id not visible
        # from the user
        # This id is used by _create_set_operation_menu
        # Keys: user text; values: set operators
        # See menu_setup()
        self.set_operations_mapping = dict()

        # call on_current_row_changed when item selection changed
        self.view.selectionModel().currentRowChanged.connect(
            self.on_current_row_changed
        )
        self.toolbar.addAction(FIcon(0xF0453), self.tr("Reload"), self.load)

    def on_open_project(self, conn):
        self.model.conn = conn
        self.conn = conn
        self.on_refresh()

    def on_refresh(self):
        self.view.selectionModel().blockSignals(True)
        self.model.load()
        self.source = self.mainwindow.state.source
        model_index = self.model.find_record(self.source)
        self.view.setCurrentIndex(model_index)
        self.view.selectionModel().blockSignals(False)

    @Slot()
    def on_current_row_changed(self):
        """This methods trigger the signal for the view
        Note:
            I don't broadcast the signal rowChanged to selectionChanged directly
            because I need to block signals for the view only
        """

        index = self.view.currentIndex()
        source = self.model.record(index)["name"]

        self.mainwindow.state.source = source
        self.mainwindow.refresh_plugins(sender=self)

    def menu_setup(self, locked_selection=False):
        """Setup popup menu
        :key locked_selection: Allow to mask edit/remove actions (default False)
            Used on special selections like the default one (named variants).
        :type locked_selection: <boolean>
        """
        menu = QMenu()

        if not locked_selection:
            menu.addAction(FIcon(0xF0900), self.tr("Edit"), self.edit_selection)

        #  Create action for bed
        menu.addAction(
            FIcon(0xF0219),
            self.tr("Intersect with BED file ..."),
            self.create_selection_from_bed,
        )

        # Set operations on selections: create mapping and actions
        set_icons_ids = (0xF0779, 0xF077C, 0xF0778)
        set_texts = (self.tr("Intersect"), self.tr("Difference"), self.tr("Union"))
        set_internal_ids = ("&", "-", "|")
        # Map the operations with an internal id not visible for the user
        # This id is used by _create_set_operation_menu
        # Keys: user text; values: internal ids
        self.set_operations_mapping = dict(zip(set_texts, set_internal_ids))

        # Create actions
        [
            menu.addMenu(self._create_set_operation_menu(FIcon(icon_id), text))
            for icon_id, text in zip(set_icons_ids, set_texts)
        ]

        if not locked_selection:
            menu.addSeparator()
            menu.addAction(FIcon(0xF0A7A), self.tr("Remove"), self.remove_selection)
        return menu

    def load(self):
        """Load selection model and update the view"""
        # Block signals during the insertions
        self.is_loading = True
        self.model.load()

        # Select record according to query.selection
        current_index = 0
        if self.source:
            current_index = self.model.find_record(self.source)
        self.view.setCurrentIndex(current_index)

        self.is_loading = False

    def ask_and_check_selection_name(self):
        """Get from the user a selection name validated or None

        TODO : create a sql.selection_exists(name) to check if selection already exists
        """
        name, success = QInputDialog.getText(
            self,
            self.tr("Type a name for selection"),
            self.tr("Selection name:"),
            QLineEdit.Normal,
        )
        if not success:
            return

        if name == DEFAULT_SELECTION_NAME:
            LOGGER.error(
                "SourceEditorWidget:save_current_query:: '%s' is a reserved name for a selection.",
                name,
            )
            self.mainwindow.status_bar.showMessage(
                self.tr("'%s' is a reserved name for a selection!") % name
            )
            QMessageBox.critical(
                self,
                self.tr("Error while creating the selection"),
                self.tr("'%s' is a reserved name for a selection!") % name,
            )
        elif name in {record["name"] for record in self.model.records}:
            LOGGER.error(
                "SourceEditorWidget:save_current_query:: '%s' is already used!", name
            )
            self.mainwindow.status_bar.showMessage(
                self.tr("'%s' is already used for a selection!") % name
            )
            QMessageBox.critical(
                self,
                self.tr("Error while creating the selection"),
                self.tr("'%s' is already used for a selection!") % name,
            )
        else:
            return name

    # def save_current_query(self):
    #     """Open a dialog box to save the current query into a selection
    #     TODO: désactivé?
    #     """
    #     selection_name = self.ask_and_check_selection_name()
    #     if selection_name:
    #         self.model.save_current_query(selection_name)

    def contextMenuEvent(self, event: QContextMenuEvent):
        """Overrided: Show popup menu at the cursor position"""
        if not self.model.records:
            return
        # Detect locked selection to mask edit/remove actions
        locked_selection = False
        current_index = self.view.selectionModel().currentIndex()
        record = self.model.record(current_index)
        if record["name"] == DEFAULT_SELECTION_NAME:
            locked_selection = True
        # Show the menu
        menu = self.menu_setup(locked_selection)
        menu.exec_(event.globalPos())

    def _create_set_operation_menu(self, icon, menu_name):
        """Dynamically add submenu with the given name to popup menu"""
        menu = QMenu(menu_name)
        menu.setIcon(icon)

        # Get all the names of selections except the current one
        current_selection_name = self.model.record(self.view.currentIndex())["name"]
        records_names = set(record["name"] for record in self.model.records)
        records_names.remove(current_selection_name)

        menu.addSeparator()
        # Create an action in the menu for all the remaining elements
        for record_name in records_names:
            action = menu.addAction(record_name)
            # This hidden data is used by _make_set_operation() to detect
            # which operation to do
            action.setData(self.set_operations_mapping[menu_name])
            action.triggered.connect(self._make_set_operation)

        return menu

    def _make_set_operation(self):
        """Do set operation
        .. note:: Called when a "set" submenu of the popup menu is triggered.
        """
        action = self.sender()

        selection_name = self.ask_and_check_selection_name()
        if not selection_name:
            return

        # Get the action's internal data to know which set operation to do
        # See setData()
        set_operator = action.data()

        # Get the records and extract their database id to build 2 Selections objects
        # {'id': 2, 'name': 'dqzdezd', 'count': 3, 'query': 'SELECT variants.id ...}
        # {'id': 1, 'name': 'variants', 'count': 11, 'query': ''}
        record_1 = self.model.record(self.view.selectionModel().currentIndex())
        record_2 = self.model.record(self.model.find_record(action.text()))

        ret = command.set_cmd(
            self.model.conn,
            selection_name,
            record_1["name"],
            record_2["name"],
            set_operator,
        )
        if not ret:
            QMessageBox.critical(
                self,
                self.tr("Error while creating the selection"),
                self.tr("Error while creating the selection, please check the logs"),
            )
            self.mainwindow.status_bar.showMessage(
                self.tr("Fail to create the selection!")
            )
            return

        self.load()

    def remove_selection(self):
        """Remove a selection from the database"""
        msg = QMessageBox(icon=QMessageBox.Warning)
        msg.setText(self.tr("Are you sure you want to remove this selection?"))
        msg.setStandardButtons(QMessageBox.Yes | QMessageBox.No)

        if msg.exec_() == QMessageBox.Yes:
            for index in self.view.selectionModel().selectedRows():
                self.model.remove_record(index)

            # Reload the UI, otherwise, the old selection is still in popup menu
            self.load()

    def edit_selection(self):
        """Update a selection in the database
        .. note:: We do not reload all the UI for this
        """
        current_index = self.view.selectionModel().currentIndex()

        selection_name = self.ask_and_check_selection_name()
        if current_index and selection_name:
            old_record = self.model.record(current_index)
            old_record["name"] = selection_name
            self.model.edit_record(current_index, old_record)

    def create_selection_from_bed(self):
        """Ask user for a bed file and create a new selection from it """
        # Reload last directory used
        app_settings = QSettings()
        last_directory = app_settings.value("last_directory", QDir.homePath())

        filepath, _ = QFileDialog.getOpenFileName(
            self,
            self.tr("Open BED file"),
            last_directory,
            self.tr("BED - Browser Extensible Data (*.bed)"),
        )
        if not filepath:
            return

        selection_name = self.ask_and_check_selection_name()
        if not selection_name:
            return

        current_index = self.view.selectionModel().currentIndex()
        current_selection = self.model.record(current_index)
        source = current_selection["name"]

        # Open bed intervals & create selection
        intervals = BedReader(filepath)
        sql.create_selection_from_bed(
            self.model.conn, source, selection_name, intervals
        )
        # Refresh UI
        self.model.load()


if __name__ == "__main__":

    import sqlite3
    import sys
    from PySide2.QtWidgets import QApplication

    app = QApplication(sys.argv)

    view = SourceEditorWidget(conn=sqlite3.connect("examples/test.snpeff.vcf.db"))
    view.show()

    app.exec_()
